#!/usr/bin/env python3
# -*- coding: utf-8 -*-

from kaikout.log import Logger
import sys

logger = Logger(__name__)

class XmppPresence:


    def send(self, jid, status_message, presence_type=None, status_type=None):
        function_name = sys._getframe().f_code.co_name
        logger.debug(f'{function_name},{jid},Start')
        logger.debug(f'{function_name},{jid},{status_message}')
        jid_from = str(self.boundjid) if self.is_component else None
        self.send_presence(pto=jid,
                           pfrom=jid_from,
                           pshow=status_type,
                           pstatus=status_message,
                           ptype=presence_type)
        logger.debug(f'{function_name},{jid},Finish')


    def subscription(self, jid, presence_type):
        function_name = sys._getframe().f_code.co_name
        logger.debug(f'{function_name},{jid},Start')
        logger.debug(f'{function_name},{jid},{presence_type}')
        self.send_presence_subscription(pto=jid,
                                        pfrom=self.boundjid.bare,
                                        ptype=presence_type,
                                        pnick=self.alias)
        logger.debug(f'{function_name},{jid},Finish')



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

class Documentation:

    def about():
        return ('KaikOut'
                '\n'
                'KaikOut - Kaiko Out group chat anti-spam and moderation bot.'
                'Spam has never been in fashion.'
                '\n\n'
                'KaikOut is an XMPP bot that suprvises group chat activity '
                'and assists in blocking and preventing of unsolicited type '
                'of messages.'
                '\n\n'
                'KaikOut is a portmanteau of Kaiko and Out'
                '\n'
                'Kaiko (懐古) translates from Japanese to "Old-Fashioned"'
                '\n\n'
                'https://git.xmpp-it.net/sch/Kaikout'
                '\n\n'
                'Copyright 2024 Schimon Jehudah Zackary'
                '\n\n'
                'Made in Switzerland'
                '\n\n'
                '🇨🇭️')

    def commands():
        return ("add URL [tag1,tag2,tag3,...]"
                "\n"
                " Bookmark URL along with comma-separated tags."
                "\n\n"
                "mod name <ID> <TEXT>"
                "\n"
                " Modify bookmark title."
                "\n"
                "mod note <ID> <TEXT>"
                "\n"
                " Modify bookmark description."
                "\n"
                "tag [+|-] <ID> [tag1,tag2,tag3,...]"
                "\n"
                " Modify bookmark tags. Appends or deletes tags, if flag tag "
                "is preceded by \'+\' or \'-\' respectively."
                "\n"
                "del <ID> or <URL>"
                "\n"
                " Delete a bookmark by ID or URL."
                "\n"
                "\n"
                "id <ID>"
                "\n"
                " Print a bookmark by ID."
                "\n"
                "last"
                "\n"
                " Print most recently bookmarked item."
                "\n"
                "tag <TEXT>"
                "\n"
                " Search bookmarks of given tag."
                "\n"
                "search <TEXT>"
                "\n"
                " Search bookmarks by a given search query."
                "\n"
                "search any <TEXT>"
                "\n"
                " Search bookmarks by a any given keyword."
                # "\n"
                # "regex"
                # "\n"
                # " Search bookmarks using Regular Expression."
                "\n")

    def notice():
        return ('Copyright 2024 Schimon Jehudah Zackary'
                '\n\n'
                'Permission is hereby granted, free of charge, to any person '
                'obtaining a copy of this software and associated '
                'documentation files (the “Software”), to deal in the '
                'Software without restriction, including without limitation '
                'the rights to use, copy, modify, merge, publish, distribute, '
                'sublicense, and/or sell copies of the Software, and to '
                'permit persons to whom the Software is furnished to do so, '
                'subject to the following conditions:'
                '\n\n'
                'The above copyright notice and this permission notice shall '
                'be included in all copies or substantial portions of the '
                'Software.'
                '\n\n'
                'THE SOFTWARE IS PROVIDED “AS IS”, WITHOUT WARRANTY OF ANY '
                'KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE '
                'WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR '
                'PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR '
                'COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER '
                'LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR '
                'OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE '
                'SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.')




#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""

TODO

1) Send message to inviter that bot has joined to groupchat.

2) If groupchat requires captcha, send the consequent message.

3) If groupchat error is received, send that error message to inviter.

FIXME

1) Save name of groupchat instead of jid as name

"""
from asyncio import TimeoutError
from kaikout.log import Logger
from slixmpp.exceptions import IqError, IqTimeout, PresenceError
import sys

logger = Logger(__name__)


class XmppMuc:


    def get_affiliation(self, room, alias):
        """Get an affiliation of a specified alias"""
        function_name = sys._getframe().f_code.co_name
        logger.debug(f'{function_name},{room},Start')
        logger.debug(f'{function_name},{room},{alias}')
        affiliation = self.plugin['xep_0045'].get_jid_property(
            room, alias, 'affiliation')
        logger.debug(f'{function_name},{room},Finish')
        return affiliation


    async def get_affiliation_list(self, room, affiliation):
        """Get an affiliation list from groupchat config"""
        function_name = sys._getframe().f_code.co_name
        logger.debug(f'{function_name},{room},Start')
        logger.debug(f'{function_name},{room},{affiliation}')
        try:
            jids = await self.plugin['xep_0045'].get_affiliation_list(
                room, affiliation)
        except Exception as e:
            jids = None
            logger.error(f'{function_name},{room},KaikOut has failed to query '
                         f'the server for a list of Jabber IDs with the '
                         f'affiliation "{affiliation}" for groupchat {room}')
            logger.error(e)
        logger.debug(f'{function_name},{room},Finish')
        return jids


    def get_alias(self, room, jid):
        function_name = sys._getframe().f_code.co_name
        logger.debug(f'{function_name},{room},Start')
        logger.debug(f'{function_name},{room},{jid}')
        alias = self.plugin['xep_0045'].get_nick(room, jid)
        logger.debug(f'{function_name},{room},Finish')
        return alias


    def get_full_jid(self, room, alias):
        """
        Get full JId.
    
        Parameters
        ----------
        jid_full : str
            Full Jabber ID.
        """
        function_name = sys._getframe().f_code.co_name
        logger.debug(f'{function_name},{room},Start')
        logger.debug(f'{function_name},{room},{alias}')
        jid_full = self.plugin['xep_0045'].get_jid_property(
            room, alias, 'jid')
        logger.debug(f'{function_name},{room},Finish')
        return jid_full


    def get_joined_rooms(self):
        function_name = sys._getframe().f_code.co_name
        logger.debug(f'{function_name},,Start')
        rooms = self.plugin['xep_0045'].get_joined_rooms()
        logger.debug(f'{function_name},,Finish')
        return rooms
    
    
    def get_role(self, room, alias):
        """Get a role of a specified alias"""
        function_name = sys._getframe().f_code.co_name
        logger.debug(f'{function_name},{room},Start')
        role = self.plugin['xep_0045'].get_jid_property(
            room, alias, 'role')
        logger.debug(f'{function_name},{room},Finish')
        return role


    async def get_role_list(self, room, role):
        """Get a role list from groupchat config"""
        function_name = sys._getframe().f_code.co_name
        logger.debug(f'{function_name},{room},Start')
        try:
            jids = await self.plugin['xep_0045'].get_roles_list(
                room, role)
            return jids
        except Exception as e:
            logger.error(f'{function_name},{room},KaikOut has failed to query '
                         'the server for a list of Jabber IDs with the role '
                         f'"{role}" for groupchat {room}')
            logger.error(f'{function_name},{jid_bare},{str(e)}')
        logger.debug(f'{function_name},{room},Finish')


    def get_roster(self, room):
        function_name = sys._getframe().f_code.co_name
        logger.debug(f'{function_name},{room},Start')
        roster = self.plugin['xep_0045'].get_roster(room)
        logger.debug(f'{function_name},{room},Finish')
        return roster


    def get_self_alias(self, room):
        """Get self alias of a given group chat"""
        function_name = sys._getframe().f_code.co_name
        logger.debug(f'{function_name},{room},Start')
        jid_full = self.plugin['xep_0045'].get_our_jid_in_room(room)
        alias = jid_full.split('/')[1]
        logger.debug(f'{function_name},{room},Finish')
        return alias


    def is_admin(self, room, alias):
        """Check if given JID is an administrator"""
        function_name = sys._getframe().f_code.co_name
        logger.debug(f'{function_name},{room},Start')
        logger.debug(f'{function_name},{room},{alias}')
        affiliation = self.plugin['xep_0045'].get_jid_property(
            room, alias, 'affiliation')
        result = True if affiliation == 'admin' else False
        logger.debug(f'{function_name},{room},Finish')
        return result


    def is_moderator(self, room, alias):
        """Check if given JID is a moderator"""
        function_name = sys._getframe().f_code.co_name
        logger.debug(f'{function_name},{room},Start')
        logger.debug(f'{function_name},{room},{alias}')
        role = self.plugin['xep_0045'].get_jid_property(
            room, alias, 'role')
        result = True if role == 'moderator' else False
        logger.debug(f'{function_name},{room},Finish')
        return result


    # NOTE Would this properly work when Alias and Local differ?
    def is_member(self, jid_bare, jid_full):
        """Check if given JID is a member"""
        function_name = sys._getframe().f_code.co_name
        logger.debug(f'{function_name},{jid_bare},Start')
        logger.debug(f'{function_name},{jid_bare},{jid_full}')
        alias = jid_full[jid_full.index('/')+1:]
        affiliation = self.plugin['xep_0045'].get_jid_property(jid_bare, alias, 'affiliation')
        result = True if affiliation == 'member' else False
        logger.debug(f'{function_name},{jid_bare},Finish')
        return result


    def is_owner(self, room, alias):
        """Check if given JID is an owner"""
        function_name = sys._getframe().f_code.co_name
        logger.debug(f'{function_name},{room},Start')
        logger.debug(f'{function_name},{room},{alias}')
        affiliation = self.plugin['xep_0045'].get_jid_property(
            room, alias, 'affiliation')
        result = True if affiliation == 'owner' else False
        logger.debug(f'{function_name},{room},Finish')
        return result


    async def join_to_muc(self, jid_bare, alias=None, password=None):
        function_name = sys._getframe().f_code.co_name
        logger.debug(f'{function_name},{jid_bare},Start')
        if not alias: alias = self.alias
        try:
            await self.plugin['xep_0045'].join_muc_wait(
                jid_bare, alias, password=password, maxchars=0,
                maxstanzas=0, seconds=0, since=0, timeout=30)
                #presence_options = {"pfrom" : jid_from},
            result = 'success'
        except IqError as e:
            logger.error(f'{function_name},{jid_bare},Error XmppIQ')
            logger.error(f'{function_name},{jid_bare},{str(e)}')
            result = 'error'
        except IqTimeout as e:
            logger.error(f'{function_name},{jid_bare},Timeout XmppIQ')
            logger.error(f'{function_name},{jid_bare},{str(e)}')
            result = 'timeout'
        except TimeoutError as e:
            logger.error(f'{function_name},{jid_bare},Timeout AsyncIO')
            logger.error(f'{function_name},{jid_bare},{str(e)}')
            result = 'timeout'
        except PresenceError as e:
            logger.error(f'{function_name},{jid_bare},Error Presence')
            logger.error(f'{function_name},{jid_bare},{str(e)}')
            if (e.condition == 'forbidden' and
                e.presence['error']['code'] == '403'):
                logger.warning(f'{self.alias} is banned from {jid_bare}')
                result = 'ban'
            elif e.condition == 'conflict':
                logger.warning(f'{function_name},{jid_bare},{e.presence["error"]["text"]}')
                result = 'conflict'
            else:
                result = 'error'
        except Exception as e:
            logger.error(f'{function_name},{jid_bare},Unknown error')
            logger.error(f'{function_name},{jid_bare},{str(e)}')
            result = 'unknown'
        logger.debug(f'{function_name},{jid_bare},Finish')
        return result


    def leave(self, jid_bare):
        function_name = sys._getframe().f_code.co_name
        logger.debug(f'{function_name},{jid_bare},Start')
        jid_bare_self = self.boundjid.bare if self.is_component else None
        message = ('This news bot will now leave this groupchat.\n'
                   f'The JID of this news bot is xmpp:{jid_bare_self}?message')
        status_message = ('This bot has left the group. '
                         f'It can be reached directly via {jid_bare_self}')
        self.send_message(mto=jid_bare, mfrom=self.boundjid.full,
                          mbody=message, mtype='groupchat')
        self.plugin['xep_0045'].leave_muc(
            jid_bare, self.alias, status_message, jid_bare_self)
        logger.debug(f'{function_name},{jid_bare},Finish')


    async def set_affiliation(
        self, room, affiliation, jid=None, alias=None, reason=None):
        function_name = sys._getframe().f_code.co_name
        logger.debug(f'{function_name},{room},Start')
        try:
            await self.plugin['xep_0045'].set_affiliation(
                room, affiliation, jid=jid, nick=alias,
                reason=reason, ifrom=self.boundjid.bare)
        except IqError as e:
            logger.error(f'{function_name},{room},Error XmppIQ')
            logger.error(f'{function_name},{room},Could not set affiliation at room: {room}')
            logger.error(f'{function_name},{room},{str(e)}')
        except Exception as e:
            logger.error(f'{function_name},{room},Unknown error')
            logger.error(f'{function_name},{room},Could not set affiliation at room: {room}')
            logger.error(f'{function_name},{room},{str(e)}')
        logger.debug(f'{function_name},{room},Finish')


    async def set_role(self, room, alias, role, reason=None):
        function_name = sys._getframe().f_code.co_name
        logger.debug(f'{function_name},{room},Start')
        try:
            await self.plugin['xep_0045'].set_role(
                room, alias, role, reason=None, ifrom=self.boundjid.bare)
        except IqError as e:
            logger.error(f'{function_name},{room},Error XmppIQ')
            logger.error(f'{function_name},{room},Could not set role of alias: {alias}')
            logger.error(f'{function_name},{room},{str(e)}')
        except Exception as e:
            logger.error(f'{function_name},{room},Unknown error')
            logger.error(f'{function_name},{room},Could not set role of alias: {alias}')
            logger.error(f'{function_name},{room},{str(e)}')
        logger.debug(f'{function_name},{room},Finish')



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import asyncio
from datetime import datetime
from functools import partial
from kaikout.about import Documentation
from kaikout.database import DatabaseToml, Record
from kaikout.log import Logger
from kaikout.utilities import Config, BlockList, Toml
from kaikout.xmpp.adhoc import XmppAdHoc
from kaikout.xmpp.chat import XmppChat
from kaikout.xmpp.commands import XmppCommands
from kaikout.xmpp.groupchat import XmppGroupchat
from kaikout.xmpp.message import XmppMessage
from kaikout.xmpp.muc import XmppMuc
from kaikout.xmpp.observation import XmppObservation
from kaikout.xmpp.pubsub import XmppPubsub
from kaikout.xmpp.status import XmppStatus
import os
import slixmpp
import sys
import time

logger = Logger(__name__)

class XmppEvents:


    def muc_online(self, presence):
        """
        Process a presence stanza from a chat room. In this case,
        presences from users that have just come online are
        handled by sending a welcome message that includes
        the user's nickname and role in the room.

        Arguments:
            presence -- The received presence stanza. See the
                        documentation for the Presence stanza
                        to see how else it may be used.
        """
        presence_muc_nick = presence['muc']['nick']
        function_name = sys._getframe().f_code.co_name
        logger.debug(f'{function_name},{presence_muc_nick},Start')
        if presence_muc_nick != self.alias:
            presence_muc_role = presence['muc']['role']
            self.send_message(mto=presence['from'].bare,
                              mbody=f'Greetings, {presence_muc_role} {presence_muc_nick}',
                              mtype='groupchat')
        logger.debug(f'{function_name},{presence_muc_nick},Finish')


    async def on_disco_info(self, DiscoInfo):
        jid = DiscoInfo['from']
        jid_bare = jid.bare
        function_name = sys._getframe().f_code.co_name
        logger.debug(f'{function_name},{jid_bare},Start')
        await self['xep_0115'].update_caps(jid=jid)
        logger.debug(f'{function_name},{jid_bare},Finish')


    # TODO Test
    async def on_groupchat_invite(self, message):
        jid = message['from']
        jid_bare = jid.bare
        #jid_full = jid.full
        function_name = sys._getframe().f_code.co_name
        logger.debug(f'{function_name},{jid_bare},Start')
        room = message['groupchat_invite']['jid']
        result = await XmppGroupchat.join_to_groupchat(self, room)
        if result != 'ban':
            #self.bookmarks.append({'jid' : room, 'lang' : '', 'pass' : ''})
            if room not in self.bookmarks: self.bookmarks.append(room)
            Toml.save_file(self.filename_bookmarks, self.data_bookmarks)
            XmppStatus.send_status_message(self, room)
            self.add_event_handler(f'muc::{room}::got_online',
                partial(XmppEvents.on_muc_got_online, self))
            self.add_event_handler(f'muc::{room}::presence',
                partial(XmppEvents.on_muc_presence, self))
            self.add_event_handler(f'muc::{room}::self-presence',
                partial(XmppEvents.on_muc_self_presence, self))


    async def on_groupchat_direct_invite(self, message):
        room = message['groupchat_invite']['jid']
        function_name = sys._getframe().f_code.co_name
        logger.debug(f'{function_name},{room},Start')
        result = await XmppGroupchat.join_to_groupchat(self, room)
        if result != 'ban':
            #self.bookmarks.append({'jid' : room, 'lang' : '', 'pass' : ''})
            if room not in self.bookmarks: self.bookmarks.append(room)
            Toml.save_file(self.filename_bookmarks, self.data_bookmarks)
            XmppStatus.send_status_message(self, room)
            self.add_event_handler(f'muc::{room}::got_online',
                partial(XmppEvents.on_muc_got_online, self))
            self.add_event_handler(f'muc::{room}::presence',
                partial(XmppEvents.on_muc_presence, self))
            self.add_event_handler(f'muc::{room}::self-presence',
                partial(XmppEvents.on_muc_self_presence, self))
        logger.debug(f'{function_name},{room},Finish')


    async def on_message(self, message):
        function_name = sys._getframe().f_code.co_name
        logger.debug(f'{function_name},,Start')
        await XmppChat.process_message(self, message)
        # if message['type'] == 'groupchat':
        # if 'mucroom' in message.keys():
        if message.get_plugin('muc', check=True):
            room = message['muc']['room']
            alias = message['muc']['nick']
            message_body = message['body']
            identifier = message['id']
            lang = message['lang']
            timestamp_iso = datetime.now().isoformat()
            fields = ['message', timestamp_iso, alias, message_body, lang, identifier]
            filename = datetime.today().strftime('%Y-%m-%d') + '_' + room
            Record.csv(filename, fields)
            db_file = DatabaseToml.instantiate(self, room)
            timestamp = time.time()
            jid_full = XmppMuc.get_full_jid(self, room, alias)
            if jid_full and '/' in jid_full:
                jid_bare = jid_full.split('/')[0]
                XmppCommands.update_last_activity(
                    self, room, jid_bare, db_file, timestamp)
                # DatabaseToml.load_jid_settings(self, room)
                # await XmppChat.process_message(self, message)
                if (XmppMuc.is_moderator(self, room, self.alias) and
                    self.settings[room]['enabled'] and
                    alias != self.alias and
                    jid_bare and
                    jid_bare not in self.settings[room]['jid_whitelist']):
                    identifier = message['id']
                    fields = [alias, message_body, identifier, timestamp]
                    Record.toml(self, room, fields, 'message')
                    # Check for message
                    await XmppObservation.observe_message(
                        self, db_file, alias, message_body, room)
                    # Check for inactivity
                    await XmppObservation.observe_inactivity(
                        self, db_file, room)
        logger.debug(f'{function_name},,Finish')


    async def on_muc_got_online(self, presence):
        room = presence['muc']['room']
        function_name = sys._getframe().f_code.co_name
        logger.debug(f'{function_name},{room},Start')
        alias = presence['muc']['nick']
        presence_body = 'User has joined to the groupchat.'
        identifier = presence['id']
        lang = presence['lang']
        timestamp_iso = datetime.now().isoformat()
        fields = ['message', timestamp_iso, alias, presence_body, lang, identifier]
        filename = datetime.today().strftime('%Y-%m-%d') + '_' + room
        Record.csv(filename, fields)
        jid_bare = presence['muc']['jid'].bare
        fields = [jid_bare, alias, timestamp_iso]
        if jid_bare:
            if not Record.jid_exist(room, fields):
                Record.csv_jid(room, fields)
            elif not Record.alias_jid_exist(room, fields):
                Record.csv_jid(room, fields)
            if jid_bare not in self.settings[room]['last_activity']:
                message_body = (f'Welcome, {alias}! We hope that you would '
                                f'have a good time at group chat {room}.')
                # NOTE Check to which resource the message is sent.
                jid_full = XmppMuc.get_full_jid(self, room, alias)
                XmppMessage.send(self, jid_full, message_body, 'chat')
        if (XmppMuc.is_moderator(self, room, self.alias) and
            self.settings[room]['enabled'] and
            jid_bare and
            jid_bare not in self.settings[room]['jid_whitelist']):
            await XmppObservation.observe_jid(self, alias, jid_bare, room)
            # message_body = 'Greetings {} and welcome to groupchat {}'.format(alias, room)
            # XmppMessage.send(self, jid.bare, message_body, 'chat')
            # Send MUC-PM in case there is no indication for reception of 1:1
            # jid_from = presence['from']
            # XmppMessage.send(self, jid_from, message_body, 'chat')
        logger.debug(f'{function_name},{room},Finish')


    async def on_muc_presence(self, presence):
        room = presence['muc']['room']
        function_name = sys._getframe().f_code.co_name
        logger.debug(f'{function_name},{room},Start')
        alias = presence['muc']['nick']
        identifier = presence['id']
        jid_bare = presence['muc']['jid'].bare
        lang = presence['lang']
        status_codes = presence['muc']['status_codes']
        actor_alias = presence['muc']['item']['actor']['nick']
        # Leave group chat upon code 301 (ban) or 307 (kick).
        if jid_bare == self.boundjid.bare and (
            301 in status_codes or 307 in status_codes):
            await XmppCommands.muc_leave(self, room)
            presence_body = (
                'KaikOut has left the group chat. Reason: Received '
                f'status code {str(status_codes)} by {actor_alias}')
        elif 301 in status_codes:
            presence_body = f'User has been banned by {actor_alias}'
        elif 307 in status_codes:
            presence_body = f'User has been kicked by {actor_alias}'
        else:
            presence_body = presence['status']
        timestamp_iso = datetime.now().isoformat()
        fields = ['presence', timestamp_iso, alias, presence_body, lang, identifier]
        filename = datetime.today().strftime('%Y-%m-%d') + '_' + room
        # if identifier and presence_body:
        Record.csv(filename, fields)
        db_file = DatabaseToml.instantiate(self, room)
        if (XmppMuc.is_moderator(self, room, self.alias) and
            self.settings[room]['enabled'] and
            alias != self.alias):
            timestamp = time.time()
            fields = [alias, presence_body, identifier, timestamp]
            Record.toml(self, room, fields, 'presence')
            # Count bans and kicks
            await XmppObservation.observe_strikes(self, db_file, presence, room)
            if jid_bare and jid_bare not in self.settings[room]['jid_whitelist']:
                # Check for status message
                await XmppObservation.observe_status_message(
                    self, alias, db_file, jid_bare, presence_body, room)
                # Check for inactivity
                await XmppObservation.observe_inactivity(self, db_file, room)
        logger.debug(f'{function_name},{room},Finish')


    def on_muc_self_presence(self, presence):
        room = presence['muc']['room']
        function_name = sys._getframe().f_code.co_name
        logger.debug(f'{function_name},{room},Start')
        actor = presence['muc']['item']['actor']['nick']
        alias = presence['muc']['nick']
        if actor and alias == self.alias:
            XmppStatus.send_status_message(self, room)
        # TODO Check whether group chat is not anonymous
        if XmppMuc.is_moderator(self, room, self.alias):
            timestamp_iso = datetime.now().isoformat()
            db_file = DatabaseToml.instantiate(self, room)
            timestamp = time.time()
            for p_alias in XmppMuc.get_roster(self, room):
                jid_full = XmppMuc.get_full_jid(self, room, p_alias)
                # Upon changing role to moderator, the second
                # time (i.e. loop) would return values.
                if jid_full:
                    jid_bare = jid_full.split('/')[0]
                    fields = [jid_bare, p_alias, timestamp_iso]
                    if not Record.jid_exist(room, fields):
                        Record.csv_jid(room, fields)
                    elif not Record.alias_jid_exist(room, fields):
                        Record.csv_jid(room, fields)
                    # NOTE It seems that there is a difficulty to determine an
                    # alias when the same JID is conected from two different
                    # resources with two different aliases
                    #if p_alias == self.settings[room]['last_alias'][jid_bare]:
                    if jid_bare not in self.settings[room]['last_activity']:
                        XmppCommands.update_last_activity(
                            self, room, jid_bare, db_file, timestamp)
                    XmppCommands.update_last_alias(
                        self, room, jid_bare, db_file, p_alias)
        logger.debug(f'{function_name},{room},Finish')


    def on_reactions(self, message):
        room = message['muc']['room']
        function_name = sys._getframe().f_code.co_name
        logger.debug(f'{function_name},{room},Start')
        reactions = message['reactions']['values']
        alias = message['muc']['nick']
        affiliation = XmppMuc.get_affiliation(self, room, alias)
        if affiliation == 'member' and '👎' in reactions:
            message_body = '{} {} has reacted to message {} with {}'.format(
                affiliation, alias, message['id'], message['reactions']['plugin']['value'])
            message.reply(message_body).send()
            logger.info(message_body)
        logger.debug(f'{function_name},{room},Finish')


    def on_room_activity(self, presence):
        print('on_room_activity')
        print(presence)
        print('testing mix core')
        breakpoint()


    async def on_session_start(self, event):
        """
        Process the session_start event.

        Typical actions for the session_start event are
        requesting the roster and broadcasting an initial
        presence stanza.

        Arguments:
            event -- An empty dictionary. The session_start
                     event does not provide any additional
                     data.
        """
        function_name = sys._getframe().f_code.co_name
        logger.debug(f'{function_name},,Start')
        XmppAdHoc.commands(self)
        # await self.get_roster()
        rtbl_sources = self.data_rtbl['sources']
        for source in rtbl_sources:
            jabber_id = source['jabber_id']
            node_id = source['node_id']
            subscribe = await XmppPubsub.subscribe(self, jabber_id, node_id)
            if subscribe['pubsub']['subscription']['subscription'] == 'subscribed':
                rtbl_list = await XmppPubsub.get_items(self, jabber_id, node_id)
                rtbl_items = rtbl_list['pubsub']['items']
                muc_bans_sha256 = self.blocklist['entries']['xmppbl.org']['muc_bans_sha256']
                for item in rtbl_items:
                    item_id = item['id']
                    if item_id not in muc_bans_sha256: muc_bans_sha256.append(item_id)
                    # TODO Extract items item_payload.find(namespace + 'title')
                    # NOTE (Pdb)
                    # for i in item['payload'].iter(): i.attrib
                    # {'reason': 'urn:xmpp:reporting:abuse'}
                self.blocklist['entries']['xmppbl.org']['muc_bans_sha256'] = muc_bans_sha256
                Toml.save_file(self.filename_blocklist, self.blocklist)
                del self.filename_blocklist
#           subscribe['from'] = xmppbl.org
#           subscribe['pubsub']['subscription']['node'] = 'muc_bans_sha256'
            subscriptions = await XmppPubsub.get_node_subscriptions(
                self, jabber_id, node_id)
        await self['xep_0115'].update_caps()
        rooms = await XmppGroupchat.autojoin_to_groupchat(self)
        # See also get_joined_rooms of slixmpp.plugins.xep_0045
        for room in rooms:
            XmppStatus.send_status_message(self, room)
            self.add_event_handler(f'muc::{room}::got_online',
                partial(XmppEvents.on_muc_got_online, self))
            self.add_event_handler(f'muc::{room}::presence',
                partial(XmppEvents.on_muc_presence, self))
            self.add_event_handler(f'muc::{room}::self-presence',
                partial(XmppEvents.on_muc_self_presence, self))
        await asyncio.sleep(5)
        self.send_presence(
            #pshow='available',
            pstatus='🛡 KaikOut Moderation Bot')
        logger.debug(f'{function_name},,Finish')



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

from kaikout.database import DatabaseToml
from kaikout.log import Logger
from kaikout.xmpp.commands import XmppCommands

logger = Logger(__name__)

class XmppModeration:


    def moderate_last_activity(self, room, jid_bare, timestamp):
        result = None
        if 'last_activity' not in self.settings[room]:
            self.settings[room]['last_activity'] = {}
        if jid_bare not in self.settings[room]['last_activity']:
            # self.settings[room]['last_activity'][jid_bare] = timestamp
            # last_activity_for_jid = self.settings[room]['last_activity'][jid_bare]
            db_file = DatabaseToml.instantiate(self, room)
            # DatabaseToml.update_jid_settings(self, room, db_file, 'last_activity', last_activity_for_jid)
            XmppCommands.update_last_activity(self, room, jid_bare, db_file, timestamp)
        else:
            jid_last_activity = self.settings[room]['last_activity'][jid_bare]
            span_inactivity = self.settings[room]['inactivity_span']
            span_inactivity_in_seconds = 60 * 60 * 24 * span_inactivity
            if timestamp - jid_last_activity > span_inactivity_in_seconds:
                result = 'Inactivity'
                return result, span_inactivity
            span_inactivity_warn = self.settings[room]['inactivity_warn']
            span_inactivity_warn_in_seconds = 60 * span_inactivity_warn
            if timestamp - jid_last_activity > span_inactivity_in_seconds - span_inactivity_warn_in_seconds:
                time_left = span_inactivity_in_seconds - (timestamp - jid_last_activity)
                time_left_in_seconds = time_left / 60 / 60
                result = 'Warning'
                return result, time_left_in_seconds
        if not result:
            return None, None


    # TODO Consider this repetition.
    # Alias1: Greetings.
    # Alias2: Greetings.
    # Alias3: Greetings.
    # Alias4: Greetings.
    # Alias5: Greetings.
    def moderate_message(self, message_body, room):
        # Process a message for faults and unsolicitations.
        message_body_lower = message_body.lower()
        message_body_lower_split = message_body_lower.split()

        # Count instances.
        message_body_lower_split_sort = list(set(message_body_lower_split))
        for i in message_body_lower_split_sort:
            count = message_body_lower_split.count(i)
            count_max = self.settings[room]['count']
            if count > count_max:
                reason = 'Spam (Repetition)'
                return reason

        # Check for denied phrases or words.
        blocklist = self.settings[room]['deny']
        for i in blocklist:
            if i in message_body_lower:
                # jid_bare = XmppMuc.get_full_jid(self, room, alias).split('/')[0]
                reason = 'Spam (Blocklist)'
                return reason

        # Check for frequency.
        message_activity = self.settings[room]['activity_message']
        length = len(message_activity)
        message_frequency = self.settings[room]['frequency_messages']
        if (message_activity[length-1]['timestamp'] - message_activity[length-2]['timestamp'] < message_frequency and
            message_activity[length-1]['alias'] == message_activity[length-2]['alias'] and
            message_activity[length-1]['id'] != message_activity[length-2]['id']):
            reason = 'Spam (Flooding)'
            return reason
        
        # message_activity = self.settings[room]['activity_message']
        # count = {}
        # for i in message_activity:
        #     message_activity.pop()
        #     for j in message_activity:
        #         if (i['timestamp'] - j['timestamp'] < 3 and
        #             i['alias'] == j['alias'] and
        #             i['id'] != j['id']):
        #             if 'alias' in count:
        #                 count['alias'] += 1
        #             else:
        #                 count['alias'] = 0


    def moderate_status_message(self, status_message, room):
        # Process a status message for faults and unsolicitations.
        status_message_lower = status_message.lower()
        status_message_lower_split = status_message_lower.split()
        reason = None

        # Count instances.
        message_body_lower_split_sort = list(set(status_message_lower_split))
        for i in message_body_lower_split_sort:
            count = status_message_lower_split.count(i)
            count_max = self.settings[room]['count']
            if count > count_max:
                reason = 'Spam (Repetition)'
                return reason, 1

        # Check for denied phrases or words.
        blocklist = self.settings[room]['deny']
        for i in blocklist:
            if i in status_message_lower:
                # jid_bare = XmppMuc.get_full_jid(self, room, alias).split('/')[0]
                reason = 'Spam (Blocklist)'
                return reason, 1

        # Check for frequency.
        presence_activity = self.settings[room]['activity_presence']
        length = len(presence_activity)
        presence_frequency = self.settings[room]['frequency_presence']
        if (presence_activity[length-1]['body'] and
            presence_activity[length-2]['body'] and
            presence_activity[length-1]['timestamp'] - presence_activity[length-2]['timestamp'] < presence_frequency and
            presence_activity[length-1]['alias'] == presence_activity[length-2]['alias'] and
            presence_activity[length-1]['id'] != presence_activity[length-2]['id'] and
            presence_activity[length-1]['body'] != presence_activity[length-2]['body']):
            reason = 'Spam (Flooding)'
            return reason, 0

        if not reason:
            return None, None



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

from email.utils import parseaddr
import hashlib
from kaikout.log import Logger
#import kaikout.sqlite as sqlite
import os
import sys
import tomli_w
from urllib.parse import urlsplit

try:
    import tomllib
except:
    import tomli as tomllib

logger = Logger(__name__)


class Config:


    def get_default_data_directory():
        """
        Determine the directory path where data files be stored.
    
        * If $XDG_DATA_HOME is defined, use it;
        * else if $HOME exists, use it;
        * else if the platform is Windows, use %APPDATA%;
        * else use the current directory.

        Returns
        -------
        str
            Path to data directory.
        """
        directory_home = os.environ.get('HOME')
        if directory_home:
            data_home = os.path.join(directory_home, '.local', 'share')
            return os.path.join(data_home, 'kaikout')
        elif sys.platform == 'win32':
            data_home = os.environ.get('APPDATA')
            if data_home is None:
                return 'kaikout_data'
        else:
            return 'kaikout_data'


    def get_default_config_directory():
        """
        Determine the directory path where configuration files be stored.
    
        * If $XDG_CONFIG_HOME is defined, use it;
        * else if $HOME exists, use it;
        * else if the platform is Windows, use %APPDATA%;
        * else use the current directory.

        Returns
        -------
        str
            Path to configuration directory.
        """
    #    directory_config_home = xdg.BaseDirectory.xdg_config_home
        directory_config_home = os.environ.get('XDG_CONFIG_HOME')
        if directory_config_home is None:
            directory_home = os.environ.get('HOME')
            if directory_home is None:
                if sys.platform == 'win32':
                    directory_config_home = os.environ.get('APPDATA')
                    if directory_config_home is None:
                        return 'kaikout_config'
                else:
                    return 'kaikout_config'
            else:
                directory_config_home = os.path.join(directory_home, '.config')
        return os.path.join(directory_config_home, 'kaikout')

    def get_default_cache_directory():
        """
        Determine the directory path where cache files be stored.
    
        * If $XDG_CACHE_HOME is defined, use it;
        * else if $HOME exists, use it;
        * else if the platform is Windows, use %APPDATA%;
        * else use the current directory.
    
        Returns
        -------
        str
            Path to cache directory.
        """
    #    cache_home = xdg.BaseDirectory.xdg_cache_home
        directory_cache_home = os.environ.get('XDG_CACHE_HOME')
        if directory_cache_home is None:
            directory_home = os.environ.get('HOME')
            if directory_home is None:
                if sys.platform == 'win32':
                    directory_cache_home = os.environ.get('APPDATA')
                    if directory_cache_home is None:
                        return 'kaikout_cache'
                else:
                    return 'kaikout_cache'
            else:
                directory_cache_home = os.path.join(directory_home, '.cache')
        return os.path.join(directory_cache_home, 'kaikout')


class Documentation:


    def manual(filename, section=None, command=None):
        function_name = sys._getframe().f_code.co_name
        logger.debug('{}: filename: {}'.format(function_name, filename))
        config_dir = Config.get_default_config_directory()
        config_file = config_dir + '/' + filename
        cmds = Toml.open_file(config_file)
        if section == 'all':
            cmd_list = ''
            for cmd in cmds:
                for i in cmds[cmd]:
                    cmd_list += cmds[cmd][i] + '\n'
        elif command and section:
            try:
                cmd_list = cmds[section][command]
            except KeyError as e:
                logger.error(e)
                cmd_list = None
        elif section:
            try:
                cmd_list = []
                for cmd in cmds[section]:
                    cmd_list.extend([cmd])
            except KeyError as e:
                logger.error('KeyError:' + str(e))
                cmd_list = None
        else:
            cmd_list = []
            for cmd in cmds:
                cmd_list.extend([cmd])
        return cmd_list


class BlockList:


    def get_filename():
        """
        Get pathname of filename.
        If filename does not exist, create it.

        Parameters
        ----------
        None.
    
        Returns
        -------
        filename : str
            Pathname.
        """
        data_dir = Config.get_default_data_directory()
        if not os.path.isdir(data_dir): os.mkdir(data_dir)
        filename = os.path.join(data_dir, r"blocklist.toml")
        if not os.path.exists(filename):
            data = {'entries' : {}}
            Toml.save_file(filename, data)
        return filename

class String:

    def md5_hash(url):
        """
        Hash URL string to MD5 checksum.

        Parameters
        ----------
        url : str
            URL.
    
        Returns
        -------
        url_digest : str
            Hashed URL as an MD5 checksum.
        """
        url_encoded = url.encode()
        url_hashed = hashlib.md5(url_encoded)
        url_digest = url_hashed.hexdigest()
        return url_digest

class Toml:

    def open_file(filename: str) -> dict:
        with open(filename, mode="rb") as fn:
            data = tomllib.load(fn)
            return data

    def save_file(filename: str, data: dict) -> None:
        with open(filename, 'w') as fn:
            data_as_string = tomli_w.dumps(data)
            fn.write(data_as_string)

class Url:

    def check_xmpp_uri(uri):
        """
        Check validity of XMPP URI.

        Parameters
        ----------
        uri : str
            URI.
    
        Returns
        -------
        jid : str
            JID or None.
        """
        jid = urlsplit(uri).path
        if parseaddr(jid)[1] != jid:
            jid = False
        return jid



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

from kaikout.database import DatabaseToml
from kaikout.log import Logger
from kaikout.xmpp.muc import XmppMuc
from kaikout.xmpp.presence import XmppPresence
import sys

logger = Logger(__name__)


class XmppStatus:


    def send_status_message(self, room, status_mode=None, status_text=None):
        """
        Send status message.
    
        Parameters
        ----------
        jid : str
            Jabber ID.
        """
        function_name = sys._getframe().f_code.co_name
        logger.debug(f'{function_name},{room},Start')
        if not status_mode and not status_text:
            if XmppMuc.is_moderator(self, room, self.alias):
                if room not in self.settings:
                    DatabaseToml.instantiate(self, room)
                    # DatabaseToml.load_jid_settings(self, room)
                if self.settings[room]['enabled']:
                    jid_task = self.actions[room] if room in self.actions else None
                    if jid_task and len(jid_task):
                        status_mode = 'dnd'
                        status_text = jid_task[list(jid_task.keys())[0]]
                    else:
                        status_mode = 'available'
                        status_text = '👁️ Moderating'
                else:
                    status_mode = 'xa'
                    status_text = '💤 Disabled'
            else:
                status_text = '⚠️ KaikOut requires moderation privileges'
                status_mode = 'away'
        XmppPresence.send(self, room, status_text, status_type=status_mode)
        logger.debug(f'{function_name},{room},Finish')



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import asyncio
from hashlib import sha256
from kaikout.database import DatabaseToml
from kaikout.log import Logger
from kaikout.xmpp.commands import XmppCommands
from kaikout.xmpp.message import XmppMessage
from kaikout.xmpp.moderation import XmppModeration
from kaikout.xmpp.muc import XmppMuc
import sys

logger = Logger(__name__)

class XmppObservation:


    async def observe_inactivity(self, db_file, room):
        function_name = sys._getframe().f_code.co_name
        logger.debug(f'{function_name},{room},Start')
        logger.debug(f'{function_name},{room},{db_file}')
        # Check for inactivity
        if self.settings[room]['check_inactivity']:
            roster_muc = XmppMuc.get_roster(self, room)
            for alias in roster_muc:
                if alias != self.alias:
                    jid_full = XmppMuc.get_full_jid(self, room, alias)
                    jid_bare = jid_full.split('/')[0]
                    result, span = XmppModeration.moderate_last_activity(
                        self, room, jid_bare, timestamp)
                    if result:
                        message_to_participant = None
                        if 'inactivity_notice' not in self.settings[room]:
                            self.settings[room]['inactivity_notice'] = []
                        noticed_jids = self.settings[room]['inactivity_notice']
                        if result == 'Inactivity':
                            if jid_bare in noticed_jids:
                                noticed_jids.remove(jid_bare)
                            # FIXME Counting and creating of key entry "score_inactivity" appear not to occur.
                            score_inactivity = XmppCommands.raise_score_inactivity(
                                self, room, jid_bare, db_file)
                            reason = 'Inactivity detected'
                            if score_inactivity > 10:
                                jid_bare = await XmppCommands.outcast(self, room, alias, reason)
                                # admins = await XmppMuc.get_affiliation_list(self, room, 'admin')
                                # owners = await XmppMuc.get_affiliation_list(self, room, 'owner')
                                moderators = set(await XmppMuc.get_role_list(
                                    self, room, 'moderator'))
                                # Report to the moderators.
                                message_to_moderators = (
                                    f'Participant {alias} ({jid_bare}) has '
                                    f'been banned from group chat {room} for '
                                    'being inactive for over '
                                    f'{score_inactivity} times.')
                                for alias in moderators:
                                    # jid_full = presence['muc']['jid']
                                    jid_full = XmppMuc.get_full_jid(
                                        self, room, alias)
                                    XmppMessage.send(self, jid_full,
                                        message_to_moderators, 'chat')
                                # Inform the subject.
                                message_to_participant = (
                                    f'You were banned from groupchat {room} '
                                    'due to being inactive for over '
                                    f'{score_inactivity} times.  Please '
                                    'contact the moderators if you think this '
                                    'was a mistake')
                            else:
                                await XmppCommands.kick(self, room, alias, reason)
                                message_to_participant = (
                                    f'You were expelled from groupchat {room} '
                                    f'due to being inactive for over {span} '
                                    'days.')
                            XmppCommands.remove_last_activity(
                                self, room, jid_bare, db_file)
                        elif result == 'Warning' and jid_bare not in noticed_jids:
                            noticed_jids.append(jid_bare)
                            time_left = int(span)
                            if not time_left: time_left = 'an'
                            message_to_participant = (
                                'This is an inactivity-warning.\n'
                                'If you would continue to be idle, you will be '
                                f'expelled from group chat {room} within '
                                f'{int(span) or "an"} hour time.')
                        DatabaseToml.update_jid_settings(self, room, db_file,
                            'inactivity_notice', noticed_jids)
                        if message_to_participant:
                            XmppMessage.send(
                                self, jid_bare, message_to_participant, 'chat')
        logger.debug(f'{function_name},{room},Finish')


    async def observe_jid(self, alias, jid_bare, room):
        function_name = sys._getframe().f_code.co_name
        logger.debug(f'{function_name},{room},Start')
        logger.debug(f'{function_name},{room},JID: {jid_bare}; Alias: {alias}')
        jid_to_sha256 = sha256(jid_bare.encode('utf-8')).hexdigest()
        for jid in self.blocklist['entries']:
            if jid not in self.settings[room]['rtbl_ignore']:
                for node in self.blocklist['entries'][jid]:
                    for item_id in self.blocklist['entries'][jid][node]:
                        if jid_to_sha256 == item_id:
                            reason = (
                                'Jabber ID has been marked by RTBL of '
                                f'Publisher: {jid}; Node: {node}.')
                            await XmppCommands.devoice(
                                self, room, alias, reason)
                            break
        logger.debug(f'{function_name},{room},Finish')


    async def observe_message(self, db_file, alias, message_body, room):
        function_name = sys._getframe().f_code.co_name
        logger.debug(f'{function_name},{room},Start')
        logger.debug(f'{function_name},{room},DB: {db_file}; Alias: {alias}; Message: {message_body}')
        if self.settings[room]['check_message']:
            reason = XmppModeration.moderate_message(self, message_body, room)
            if reason:
                score_max = self.settings[room]['score_messages']
                score = XmppCommands.raise_score(
                    self, room, alias, db_file, reason)
                if score > score_max:
                    if self.settings[room]['action']:
                        jid_bare = await XmppCommands.outcast(
                            self, room, alias, reason)
                        # admins = await XmppMuc.get_affiliation_list(self, room, 'admin')
                        # owners = await XmppMuc.get_affiliation_list(self, room, 'owner')
                        moderators = set(await XmppMuc.get_role_list(
                            self, room, 'moderator'))
                        # Report to the moderators.
                        message_to_moderators = (
                            'Participant {alias} ({jid_bare}) has been banned '
                            'from group chat {room}.')
                        for alias in moderators:
                            jid_full = XmppMuc.get_full_jid(self, room, alias)
                            XmppMessage.send(
                                self, jid_full, message_to_moderators, 'chat')
                        # Inform the subject
                        message_to_participant = (
                            f'You were banned from groupchat {room}.  Please '
                            'contact the moderators if you think this was '
                            'a mistake.')
                        XmppMessage.send(
                            self, jid_bare, message_to_participant, 'chat')
                    else:
                        await XmppCommands.devoice(self, room, alias, reason)
        logger.debug(f'{function_name},{room},Finish')


    async def observe_status_message(
        self, alias, db_file, jid_bare, presence_body, room):
        function_name = sys._getframe().f_code.co_name
        logger.debug(f'{function_name},{room},Start')
        logger.debug(f'{function_name},{room},Alias: {alias}; DB: {db_file}; JID: {jid_bare}; Presence: {presence_body}')
        if self.settings[room]['check_status']:
            reason, timer = XmppModeration.moderate_status_message(
                self, presence_body, room)
            if reason and timer and not (
                room in self.tasks and
                jid_bare in self.tasks[room] and
                'countdown' in self.tasks[room][jid_bare]):
                print('reason and timer for jid: ' + jid_bare + ' at room ' + room)
                score_max = self.settings[room]['score_presence']
                score = XmppCommands.raise_score(
                    self, room, alias, db_file, reason)
                if room not in self.tasks:
                    self.tasks[room] = {}
                if jid_bare not in self.tasks[room]:
                    self.tasks[room][jid_bare] = {}
                # if 'countdown' in self.tasks[room][jid_bare]:
                #     self.tasks[room][jid_bare]['countdown'].cancel()
                if 'countdown' not in self.tasks[room][jid_bare]:
                    seconds = self.settings[room]['timer']
                    self.tasks[room][jid_bare]['countdown'] = asyncio.create_task(
                        XmppCommands.countdown(
                            self, seconds, room, alias, reason))
                message_to_participant = (
                    f'Your status message "{presence_body}" violates policies '
                    f'of group chat {room}.\n'
                    f'To avoid consequent actions, you have {seconds} seconds '
                    'to change your status message')
                XmppMessage.send(self, jid_bare, message_to_participant, 'chat')
            elif reason and not (room in self.tasks
                                    and jid_bare in self.tasks[room] and
                                    'countdown' in self.tasks[room][jid_bare]):
                print('reason for jid: ' + jid_bare + ' at room ' + room)
                score_max = self.settings[room]['score_presence']
                score = XmppCommands.raise_score(
                    self, room, alias, db_file, reason)
                if score > score_max:
                    if self.settings[room]['action']:
                        jid_bare = await XmppCommands.outcast(
                            self, room, alias, reason)
                        # admins = await XmppMuc.get_affiliation_list(self, room, 'admin')
                        # owners = await XmppMuc.get_affiliation_list(self, room, 'owner')
                        moderators = set(await XmppMuc.get_role_list(
                            self, room, 'moderator'))
                        # Report to the moderators.
                        message_to_moderators = (
                            f'Participant {alias} ({jid_bare}) has been banned '
                            f'from group chat {room}.')
                        for alias in moderators:
                            # jid_full = presence['muc']['jid']
                            jid_full = XmppMuc.get_full_jid(self, room, alias)
                            XmppMessage.send(
                                self, jid_full, message_to_moderators, 'chat')
                        # Inform the subject.
                        message_to_participant = (
                            f'You were banned from group chat {room}.  Please '
                            'contact the moderators if you think this was a '
                            'mistake.')
                        XmppMessage.send(
                            self, jid_bare, message_to_participant, 'chat')
                    else:
                        await XmppCommands.devoice(self, room, alias, reason)
            elif (room in self.tasks and
                    jid_bare in self.tasks[room] and
                    'countdown' in self.tasks[room][jid_bare]) and not reason:
                print('cancel task for jid: ' + jid_bare + ' at room ' + room)
                print(self.tasks[room][jid_bare]['countdown'])
                if self.tasks[room][jid_bare]['countdown'].cancel():
                    print(self.tasks[room][jid_bare]['countdown'])
                    message_to_participant = 'Thank you for your cooperation.'
                    XmppMessage.send(
                        self, jid_bare, message_to_participant, 'chat')
                del self.tasks[room][jid_bare]['countdown']
        logger.debug(f'{function_name},{room},Finish')


    async def observe_strikes(self, db_file, presence, room):
        function_name = sys._getframe().f_code.co_name
        logger.debug(f'{function_name},{room},Start')
        logger.debug(f'{function_name},{room},DB {db_file}')
        if self.settings[room]['check_moderation']:
            status_codes = presence['muc']['status_codes']
            if (301 in status_codes or 307 in status_codes):
                actor_jid_bare = presence['muc']['item']['actor']['jid'].bare
                actor_alias = presence['muc']['item']['actor']['nick']
                if 301 in status_codes:
                    presence_body = f'User has been banned by {actor_alias}'
                    XmppCommands.update_score_ban(
                        self, room, actor_jid_bare, db_file)
                elif 307 in status_codes:
                    presence_body = f'User has been kicked by {actor_alias}'
                    XmppCommands.update_score_kick(
                        self, room, actor_jid_bare, db_file)
                if 'score_ban' in self.settings[room] and (
                    actor_jid_bare in self.settings[room]['score_ban']):
                    score_ban = self.settings[room]['score_ban'][actor_jid_bare]
                else:
                    score_ban = 0
                if 'score_kick' in self.settings[room] and (
                    actor_jid_bare in self.settings[room]['score_kick']):
                    score_kick = self.settings[room]['score_kick'][actor_jid_bare]
                else:
                    score_kick = 0
                score_outcast = score_ban + score_kick
                if score_outcast > self.settings[room]['score_outcast']:
                    reason = 'Moderation abuse has been triggered'
                    await XmppMuc.set_affiliation(
                        self, room, 'member', jid=actor_jid_bare, reason=reason)
                    await XmppMuc.set_role(
                        self, room, actor_alias, 'participant', reason)
        logger.debug(f'{function_name},{room},Finish')



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

from kaikout.log import Logger
from kaikout.xmpp.dataforms import XmppDataForm
import sys

logger = Logger(__name__)

class XmppForms:


    def form_admin(self, iq, session):
        jid_bare = iq['from'].bare
        function_name = sys._getframe().f_code.co_name
        logger.debug(f'{function_name},{jid_bare},Start')
        text_note = 'KaikOut Administration Panel.'
        session['notes'] = [['info', text_note]]
        logger.debug(f'{function_name},{jid_bare},Finish')
        return session


    def form_config(self, iq, session):
        jid_bare = iq['from'].bare
        function_name = sys._getframe().f_code.co_name
        logger.debug(f'{function_name},{jid_bare},Start')
        text_note = 'KaikOut Configuration Panel.'
        session['notes'] = [['info', text_note]]
        logger.debug(f'{function_name},{jid_bare},Finish')
        return session


    def form_manual(self, iq, session):
        jid_bare = iq['from'].bare
        function_name = sys._getframe().f_code.co_name
        logger.debug(f'{function_name},{jid_bare},Start')
        text_note = 'KaikOut is a moderation service for XMPP.'
        session['notes'] = [['info', text_note]]
        logger.debug(f'{function_name},{jid_bare},Finish')
        return session


    def form_report(self, iq, session):
        jid_bare = iq['from'].bare
        function_name = sys._getframe().f_code.co_name
        logger.debug(f'{function_name},{jid_bare},Start')
        text_note = 'KaikOut Abuse Report System.'
        session['notes'] = [['info', text_note]]
        logger.debug(f'{function_name},{jid_bare},Finish')
        return session


    def note_about(iq, session):
        jid_bare = iq['from'].bare
        function_name = sys._getframe().f_code.co_name
        logger.debug(f'{function_name},{jid_bare},Start')
        text_note = 'KaikOut Moderation Service.'
        session['notes'] = [['info', text_note]]
        logger.debug(f'{function_name},{jid_bare},Finish')
        return session


    def note_cancel(payload, session):
        jid_bare = jid_bare
        function_name = sys._getframe().f_code.co_name
        logger.debug(f'{function_name},,Start')
        text_note = 'Operation has been cancelled.'
        session['notes'] = [['warn', text_note]]
        logger.debug(f'{function_name},{jid_bare},Finish')
        return session


    def note_error(payload, session):
        jid_bare = jid_bare
        function_name = sys._getframe().f_code.co_name
        logger.debug(f'{function_name},,Start')
        text_note = 'An error has been encountered.'
        session['notes'] = [['error', text_note]]
        logger.debug(f'{function_name},{jid_bare},Finish')
        return session




#!/usr/bin/env python3
# -*- coding: utf-8 -*-

from kaikout.log import Logger
import sys

logger = Logger(__name__)

class XmppDataForm:


    def make_form(self, title, instructions):
        function_name = sys._getframe().f_code.co_name
        logger.debug(f'{function_name},,Start')
        form = self['xep_0004'].make_form(
            ftype='form', title=title, instructions=instructions)
        logger.debug(f'{function_name},,Finish')
        return form



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

from datetime import datetime
import logging

class Logger:

    def __init__(self, name):
        self.logger = logging.getLogger(name)

    def set_level(self, loglevel):
        match loglevel:
            case 'debug':
                level = logging.DEBUG
            case 'error':
                level = logging.ERROR
            case 'info':
                level = logging.INFO
            case 'warning':
                level = logging.WARNING
            case _:
                level = logging.NOTSET

        self.logger.setLevel(level)

        console_handler = logging.StreamHandler()
        console_handler.setLevel(level)

        formatter = logging.Formatter('[%(asctime)s] %(levelname)s:%(name)s:%(message)s')
        console_handler.setFormatter(formatter)

        # Add the console handler only if there are no handlers already
        if not self.logger.handlers:
            self.logger.addHandler(console_handler)

    def set_filename(self, filename):

        file_handler = logging.FileHandler(filename)
        file_handler.setLevel(logging.DEBUG)

        formatter = logging.Formatter('%(asctime)s,%(levelname)s,%(name)s,%(message)s')
        file_handler.setFormatter(formatter)

        # Add the console handler only if there are no handlers already
        if not self.logger.handlers:
            self.logger.addHandler(file_handler)
    
    def critical(self, message):
        self.logger.critical(message)
    
    def debug(self, message):
        self.logger.debug(message)
    
    def error(self, message):
        self.logger.error(message)
    
    def info(self, message):
        self.logger.info(message)
    
    def warning(self, message):
        self.logger.warning(message)

    # def check_difference(function_name, difference):
    #     if difference > 1:
    #         Logger.warning(message)


class Message:


    def printer(text):
        now = datetime.now()
        current_time = now.strftime("%H:%M:%S")
        print('{} {}'.format(current_time, text), end='\r')



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

from kaikout.log import Logger
from kaikout.xmpp.muc import XmppMuc
from slixmpp.exceptions import IqError, IqTimeout
import sys

logger = Logger(__name__)


class XmppUtilities:


    async def is_jid_of_moderators(self, room, jid_full):
        function_name = sys._getframe().f_code.co_name
        logger.debug(f'{function_name},{room},Start')
        # try:
        jid_bare = jid_full.split('/')[0]
        is_jid_of_moderators = False
        moderators = set(await XmppMuc.get_role_list(self, room, 'moderator'))
        for alias in moderators:
            # NOTE You might want to compare jid_bare instead of jid_full
            #if XmppMuc.get_full_jid(self, room, alias) == jid_full:
            jid_full_moderator = XmppMuc.get_full_jid(self, room, alias)
            jid_bare_moderator = jid_full_moderator.split('/')[0]
            if jid_bare == jid_bare_moderator:
                is_jid_of_moderators = True
                break
        # except IqError as e:
        #     logger.error(e)
        logger.debug(f'{function_name},{room},Finish')
        return is_jid_of_moderators


    async def get_chat_type(self, jid_full):
        """
        Check chat (i.e. JID) type.

        If iqresult["disco_info"]["features"] contains XML namespace
        of 'http://jabber.org/protocol/muc', then it is a 'groupchat'.
    
        Unless it has forward slash, which would indicate that it is
        a chat which is conducted through a groupchat.
        
        Otherwise, determine type 'chat'.

        Parameters
        ----------
        jid_full : str
            Jabber ID.

        Returns
        -------
        result : str
            'chat' or 'groupchat' or 'error'.
        """
        function_name = sys._getframe().f_code.co_name
        logger.debug(f'{function_name},{jid_full},Start')
        try:
            iqresult = await self["xep_0030"].get_info(jid=jid_full)
            features = iqresult["disco_info"]["features"]
            # identity = iqresult['disco_info']['identities']
            # if 'account' in indentity:
            # if 'conference' in indentity:
            if ('http://jabber.org/protocol/muc' in features) and not ('/' in jid_full):
                result = "groupchat"
            # TODO elif <feature var='jabber:iq:gateway'/>
            # NOTE Is it needed? We do not interact with gateways or services
            else:
                result = "chat"
            logger.info(f'{function_name},{jid_full},Chat Type: {result}')
        except (IqError, IqTimeout) as e:
            logger.error(f'{function_name},{jid_full},Error XmppIQ')
            logger.error(f'{function_name},{jid_full},{str(e)}')
            logger.warning(f'{function_name},{jid_full},Chat type could not be determined.')
            result = 'error'
        # except BaseException as e:
        #     logger.error('BaseException', str(e))
        # finally:
        #     logger.info('Chat type is:', chat_type)
        logger.debug(f'{function_name},{jid_full},Finish')
        return result

    def is_access(self, jid_bare, jid_full, chat_type):
        """Determine access privilege"""
        function_name = sys._getframe().f_code.co_name
        logger.debug(f'{function_name},{jid_bare},Start')
        logger.debug(f'{function_name},{jid_bare},JID: {jid_full}; Type: {chat_type}')
        operator = XmppUtilities.is_operator(self, jid_bare)
        if operator:
            if chat_type == 'groupchat':
                if XmppMuc.is_moderator(self, jid_bare, jid_full):
                    access = True
            else:
                access = True
        else:
            access = False
        logger.debug(f'{function_name},{jid_bare},Finish')
        return access

    def is_operator(self, jid_bare):
        """Check if given JID is an operator"""
        function_name = sys._getframe().f_code.co_name
        logger.debug(f'{function_name},{jid_bare},Start')
        result = False
        for operator in self.operators:
            if jid_bare == operator['jid']:
                result = True
                # operator_name = operator['name']
                break
        logger.debug(f'{function_name},{jid_bare},Finish')
        return result



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

from functools import partial
from kaikout.about import Documentation
from kaikout.database import DatabaseToml, Record
from kaikout.log import Logger
from kaikout.utilities import Config, BlockList, Toml
from kaikout.xmpp.events import XmppEvents
import os
import slixmpp
import sys
import time

# time_now = datetime.now()
# time_now = time_now.strftime("%H:%M:%S")

# def print_time():
#     # return datetime.now().strftime("%H:%M:%S")
#     now = datetime.now()
#     current_time = now.strftime("%H:%M:%S")
#     return current_time

logger = Logger(__name__)

class XmppClient(slixmpp.ClientXMPP):

    """
    KaikOut - A moderation bot for Jabber/XMPP.
    KaikOut is a chat control bot for XMPP groupchats.
    """

    def __init__(self, jid, password, hostname, port, alias):
        slixmpp.ClientXMPP.__init__(self, jid, password, hostname, port, alias)

        # Get directories
        self.directory_config = Config.get_default_config_directory()
        self.directory_shared = Config.get_default_data_directory()

        # A handler for accounts.
        filename_accounts = os.path.join(self.directory_config, 'accounts.toml')
        self.data_accounts = Toml.open_file(filename_accounts)
        self.data_accounts_xmpp = self.data_accounts['xmpp']

        # A handler for blocklist.
        self.filename_blocklist = os.path.join(self.directory_shared, 'blocklist.toml')
        self.blocklist = Toml.open_file(self.filename_blocklist)

        # A handler for blocklist.
        filename_rtbl = os.path.join(self.directory_config, 'rtbl.toml')
        self.data_rtbl = Toml.open_file(filename_rtbl)

        # A handler for settings.
        filename_settings = os.path.join(self.directory_config, 'settings.toml')
        self.data_settings = Toml.open_file(filename_settings)

        # Handlers for action messages.
        self.actions = {}
        self.action_count = 0

        # A handler for alias.
        self.alias = alias

        # A handler for bookmarks.
        self.filename_bookmarks = os.path.join(self.directory_config, 'bookmarks.toml')
        self.data_bookmarks = Toml.open_file(self.filename_bookmarks)
        self.bookmarks = self.data_bookmarks['bookmarks']

        # A handler for configuration.
        self.defaults = self.data_settings['defaults']

        # Handlers for connectivity.
        self.connection_attempts = 0
        self.max_connection_attempts = 10
        self.task_ping_instance = {}
        self.reconnect_timeout = self.data_accounts_xmpp['settings']['reconnect_timeout']

        # A handler for operators.
        self.operators = self.data_accounts_xmpp['operators']

        # A handler for sessions.
        self.sessions = {}

        # A handler for settings.
        self.settings = {}

        # A handler for tasks.
        self.tasks = {}

        # Register plugins.
        xeps = [
            'xep_0004', # Data Forms
            'xep_0030', # Service Discovery
            'xep_0045', # Multi-User Chat
            'xep_0048', # Bookmarks
            'xep_0050', # Ad-Hoc Commands
            'xep_0060', # Publish-Subscribe
            'xep_0084', # User Avatar
            'xep_0085', # Chat State Notifications
            'xep_0115', # Entity Capabilities
            'xep_0122', # Data Forms Validation
            'xep_0199', # XMPP Ping
            'xep_0249', # Direct MUC Invitations
            'xep_0369', # Mediated Information eXchange (MIX)
            'xep_0437', # Room Activity Indicators
            'xep_0444', # Message Reactions
        ]

        for xep in xeps: self.register_plugin(xep)

        # Register events.
        events = [
#           ['chatstate_composing',
#            XmppEvents.on_chatstate_composing],
#           ['connection_failed',
#            XmppEvents.on_connection_failed],
            ['disco_info',
             XmppEvents.on_disco_info],
            ['groupchat_direct_invite',
             XmppEvents.on_groupchat_direct_invite], # XEP_0249
            ['groupchat_invite',
             XmppEvents.on_groupchat_invite], # XEP_0045
            ['message',
             XmppEvents.on_message],
            ['reactions',
             XmppEvents.on_reactions],
#           ['room_activity',
#            XmppEvents.on_room_activity],
#           ['session_resumed',
#            XmppEvents.on_session_resumed],
            ['session_start',
             XmppEvents.on_session_start],
        ]

        for event, handler in events:
            self.add_event_handler(
                event,
                partial(handler, self))

        # Connect and process.
        self.connect()
        self.loop.run_forever()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""

NOTE

The VCard XML fields that can be set are as follows:
    ‘FN’, ‘NICKNAME’, ‘URL’, ‘BDAY’, ‘ROLE’, ‘NOTE’, ‘MAILER’,
    ‘TZ’, ‘REV’, ‘UID’, ‘DESC’, ‘TITLE’, ‘PRODID’, ‘SORT-STRING’,
    ‘N’, ‘ADR’, ‘TEL’, ‘EMAIL’, ‘JABBERID’, ‘ORG’, ‘CATEGORIES’,
    ‘NOTE’, ‘PRODID’, ‘REV’, ‘SORT-STRING’, ‘SOUND’, ‘UID’, ‘URL’,
    ‘CLASS’, ‘KEY’, ‘MAILER’, ‘GEO’, ‘TITLE’, ‘ROLE’,
    ‘LOGO’, ‘AGENT’

TODO

1) Test XEP-0084.

2) Make sure to support all type of servers.
    
3) Catch IqError
    ERROR:slixmpp.basexmpp:internal-server-error: Database failure
    WARNING:slixmpp.basexmpp:You should catch IqError exceptions

"""

import glob
from kaikout.config import Config
import kaikout.config as config
from kaikout.log import Logger
from slixmpp.exceptions import IqTimeout, IqError
import os

logger = Logger(__name__)

# class XmppProfile:

async def update(self):
    """ Update profile """
    try:
        await set_vcard(self)
    except IqTimeout as e:
        logger.error('Profile vCard: Error Timeout')
        logger.error(str(e))
    except IqError as e:
            logger.error('Profile vCard: Error XmppIQ')
            logger.error(str(e))
    try:
        await set_avatar(self)
    except IqTimeout as e:
        logger.error('Profile Photo: Error Timeout')
        logger.error(str(e))
    except IqError as e:
            logger.error('Profile Photo: Error XmppIQ')
            logger.error(str(e))


async def set_avatar(self):
    config_dir = config.get_default_config_directory()
    if not os.path.isdir(config_dir):
        config_dir = '/usr/share/kaikout/'
    filename = glob.glob(config_dir + '/image.*')
    if not filename and os.path.isdir('/usr/share/kaikout/'):
        # filename = '/usr/share/kaikout/image.svg'
        filename = glob.glob('/usr/share/kaikout/image.*')
    if not filename:
        config_dir = os.path.dirname(__file__)
        config_dir = config_dir.split('/')
        config_dir.pop()
        config_dir = '/'.join(config_dir)
        filename = glob.glob(config_dir + '/assets/image.*')
    if len(filename):
        filename = filename[0]
        image_file = os.path.join(config_dir, filename)
        with open(image_file, 'rb') as avatar_file:
            avatar = avatar_file.read()
            # await self.plugin['xep_0084'].publish_avatar(avatar)
            try:
                await self.plugin['xep_0153'].set_avatar(avatar=avatar)
            except IqTimeout as e:
                logger.error('Profile Photo: Error Timeout 222')
                logger.error(str(e))
            except IqError as e:
                    logger.error('Profile Photo: Error XmppIQ 222')
                    logger.error(str(e))


def set_identity(self, category):
    """
    Identify for Service Descovery.

    Parameters
    ----------
    category : str
        "client" or "service".

    Returns
    -------
    None.

    """
    self['xep_0030'].add_identity(
        category=category,
        itype='news',
        name='kaikout',
        node=None,
        jid=self.boundjid.full,
    )


async def set_vcard(self):
    vcard = self.plugin['xep_0054'].make_vcard()
    profile = config.get_values('accounts.toml', 'xmpp')['profile']
    for key in profile:
        vcard[key] = profile[key]
    await self.plugin['xep_0054'].publish_vcard(vcard)




#!/usr/bin/env python3
# -*- coding: utf-8 -*-

from kaikout.log import Logger
import sys
import xml.sax.saxutils as saxutils

logger = Logger(__name__)

"""

NOTE

See XEP-0367: Message Attaching

"""

class XmppMessage:


    def send(self, jid_bare, message_body, chat_type):
        function_name = sys._getframe().f_code.co_name
        logger.debug(f'{function_name},{jid_bare},Start')
        logger.debug(f'{function_name},{jid_bare},Type: {chat_type}; Message: {message_body}')
        jid_full_self = self.boundjid.full if self.is_component else None
        self.send_message(mto=jid_bare,
                          mfrom=jid_full_self,
                          mbody=message_body,
                          mtype=chat_type)
        logger.debug(f'{function_name},{jid_bare},Finish')


    def send_headline(self, jid_bare, message_subject, message_body, chat_type):
        function_name = sys._getframe().f_code.co_name
        logger.debug(f'{function_name},{jid_bare},Start')
        logger.debug(f'{function_name},{jid_bare},Type: {chat_type}; Subject: {message_subject}; Message: {message_body}')
        jid_full_self = self.boundjid.full if self.is_component else None
        self.send_message(mto=jid_bare,
                          mfrom=jid_full_self,
                          # mtype='headline',
                          msubject=message_subject,
                          mbody=message_body,
                          mtype=chat_type,
                          mnick=self.alias)
        logger.debug(f'{function_name},{jid_bare},Finish')


    # NOTE We might want to add more characters
    # def escape_to_xml(raw_string):
    # escape_map = {
    #     '"' : '&quot;',
    #     "'" : '&apos;'
    # }
    # return saxutils.escape(raw_string, escape_map)
    def send_oob(self, jid_bare, url, chat_type):
        jid_full_self = self.boundjid.full if self.is_component else None
        url = saxutils.escape(url)
        # try:
        html = ('<body xmlns="http://www.w3.org/1999/xhtml">'
                f'<a href="{url}">{url}</a></body>')
        message = self.make_message(mto=jid_bare,
                                    mfrom=jid_full_self,
                                    mbody=url,
                                    mhtml=html,
                                    mtype=chat_type)
        message['oob']['url'] = url
        message.send()
        # except:
        #     logging.error('ERROR!')
        #     logging.error(jid, url, chat_type, html)


    # FIXME Solve this function
    def send_oob_reply_message(message, url, response):
        reply = message.reply(response)
        reply['oob']['url'] = url
        reply.send()


    # def send_reply(self, message, message_body):
    #     message.reply(message_body).send()


    def send_reply(self, message, response):
        function_name = sys._getframe().f_code.co_name
        jid_bare = message['from'].bare
        logger.debug(f'{function_name},{jid_bare},Start')
        logger.debug(f'{function_name},{jid_bare},To: {message["to"].bare}; Response: {response}')
        message.reply(response).send()
        logger.debug(f'{function_name},{jid_bare},Finish')



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

from functools import partial
from kaikout.log import Logger
from kaikout.xmpp.forms import XmppForms
import sys

logger = Logger(__name__)

class XmppAdHoc:


    def add_command(self, node, name, handler):
        function_name = sys._getframe().f_code.co_name
        logger.debug(f'{function_name},,Start')
        self['xep_0050'].add_command(node=node, name=name, handler=handler)
        logger.debug(f'{function_name},,Finish')


    def commands(self):
        function_name = sys._getframe().f_code.co_name
        logger.debug(f'{function_name},,Start')

        commands = [
            ['manual',
             '📕 Manual',
              XmppForms.form_manual],
            ['config',
             '🎛️ Configuration',
              XmppForms.form_config],
            ['report',
             '📝 Report',
              XmppForms.form_report],
            ['admin',
             '🗝 Administration',
              XmppForms.form_admin]
        ]

        for node, name, handler in commands:
            XmppAdHoc.add_command(
                self,
                node,
                name,
                partial(handler, self))

        XmppAdHoc.add_command(
            self,
            'about',
            '📜️ About',
            XmppForms.note_about)

#        XmppAdHoc.add_command(
#            self,
#            'about',
#            '📜️ About',
#            lambda session, iq: XmppForms.note_about(self, session, iq))

        logger.debug(f'{function_name},,Finish')



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

from kaikout.log import Logger, Message
from kaikout.utilities import Toml
from kaikout.xmpp.message import XmppMessage
from kaikout.xmpp.muc import XmppMuc
from kaikout.xmpp.status import XmppStatus
import random
import sys

logger = Logger(__name__)

class XmppGroupchat:

    async def join_to_groupchat(self, room):
        function_name = sys._getframe().f_code.co_name
        logger.debug(f'{function_name},{room},Start')
        result = await XmppMuc.join_to_muc(self, room)
        if result == 'ban':
            message_body = '{} is banned from {}'.format(self.alias, room)
            jid_bare = message['from'].bare
            XmppMessage.send(self, jid_bare, message_body, 'chat')
            logger.warning(f'{function_name},{room},{message_body}')
        elif result == 'conflict':
            while result == 'conflict':
                number = str(random.randrange(1000, 5000))
                logger.warning(f'{function_name},{room},Conflict. Atempting to join to {room} as {self.alias} #{number}')
                result = await XmppMuc.join_to_muc(self, room, f'{self.alias} #{number}')
        else:
            #self.bookmarks.append({'jid' : room, 'lang' : '', 'pass' : ''})
            if room not in self.bookmarks: self.bookmarks.append(room)
            Toml.save_file(self.filename_bookmarks, self.data_bookmarks)
        logger.debug(f'{function_name},{room},Finish')
        return result

    async def autojoin_to_groupchat(self):
        function_name = sys._getframe().f_code.co_name
        logger.debug(f'{function_name},,Start')
        mucs_joined = []
        for room in self.bookmarks:
            alias = self.alias
            #Message.printer(f'Joining to MUC {room} ...')
            result = await XmppMuc.join_to_muc(self, room)
            if result == 'ban':
                if room in self.bookmarks: self.bookmarks.remove(room)
                Toml.save_file(self.filename_bookmarks, self.data_bookmarks)
                logger.warning(f'{alias} is banned from {room}')
                logger.warning(f'Groupchat {room} has been removed from bookmarks')
            elif result == 'conflict':
                while result == 'conflict':
                    number = str(random.randrange(1000, 5000))
                    logger.warning(f'Conflict. Atempting to join to {room} as {self.alias} #{number}')
                    result = await XmppMuc.join_to_muc(self, room, f'{alias} #{number}')
            else:
                mucs_joined.append(room)
        logger.debug(f'{function_name},,Finish')
        return mucs_joined



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

from kaikout.log import Logger
import os
import sys
try:
    import tomllib
except:
    import tomli as tomllib

logger = Logger(__name__)


class Config:


    def get_default_data_directory():
        directory_home = os.environ.get('HOME')
        if directory_home:
            data_home = os.path.join(directory_home, '.local', 'share')
            return os.path.join(data_home, 'kaikout')
        elif sys.platform == 'win32':
            data_home = os.environ.get('APPDATA')
            if data_home is None:
                return 'kaikout_data'
        else:
            return 'kaikout_data'


    def get_default_config_directory():
        """
        Determine the directory path where configuration will be stored.
    
        * If $XDG_CONFIG_HOME is defined, use it;
        * else if $HOME exists, use it;
        * else if the platform is Windows, use %APPDATA%;
        * else use the current directory.

        Returns
        -------
        str
            Path to configuration directory.
        """
        # directory_config_home = xdg.BaseDirectory.xdg_config_home
        directory_config_home = os.environ.get('XDG_CONFIG_HOME')
        if directory_config_home is None:
            directory_home = os.environ.get('HOME')
            if directory_home is None:
                if sys.platform == 'win32':
                    directory_config_home = os.environ.get('APPDATA')
                    if directory_config_home is None:
                        return 'kaikout_config'
                else:
                    return 'kaikout_config'
            else:
                directory_config_home = os.path.join(directory_home, '.config')
        return os.path.join(directory_config_home, 'kaikout')


    def get_values(filename, key=None):
        config_dir = Config.get_default_config_directory()
        if not os.path.isdir(config_dir):
            config_dir = '/usr/share/kaikout/'
        if not os.path.isdir(config_dir):
            config_dir = os.path.dirname(__file__) + "/assets"
        config_file = os.path.join(config_dir, filename)
        with open(config_file, mode="rb") as defaults:
            result = tomllib.load(defaults)
        values = result[key] if key else result
        return values



#!/usr/bin/env python3

# Kaikout: Kaiko Out anti-spam bot for XMPP
# Copyright (C) 2024 Schimon Zackary
# This file is part of Kaikout.
# See the file LICENSE for copying permission.

# from kaikout.about import Documentation
from kaikout.log import Logger
from kaikout.utilities import Config, Toml
# from kaikout.xmpp.chat import XmppChat
from kaikout.xmpp.client import XmppClient
from kaikout.version import __version__
from getpass import getpass
from argparse import ArgumentParser
import os
import shutil
# import slixmpp
# import sys

def main():

    directory = os.path.dirname(__file__)

    # Copy data files
    directory_data = Config.get_default_config_directory()
    # TODO Utilize the actual data directory
    #directory_data = Config.get_default_data_directory()
    if not os.path.exists(directory_data):
        directory_assets = os.path.join(directory, 'assets')
        directory_assets_new = shutil.copytree(directory_assets, directory_data)
        print(f'Data directory {directory_assets_new} has been created and populated.')

    # Copy settings files
    directory_settings = Config.get_default_config_directory()
    if not os.path.exists(directory_settings):
        directory_configs = os.path.join(directory, 'configs')
        directory_settings_new = shutil.copytree(directory_configs, directory_settings)
        print(f'Configuration directory {directory_settings_new} has been created and populated.')

    # Setup the command line arguments.
    parser = ArgumentParser(description=XmppClient.__doc__)

    # Setup the command line arguments.
    parser.add_argument('-v', '--version', help='Print version',
                        action='version', version=__version__)
    parser.add_argument('--no-log', '--nolog', help='Do not create a log file',
                        action='store_const', dest='nolog',
                        const=True, default=False)

    # Output verbosity options.
    parser.add_argument('-d', '--debug', help='set logging to DEBUG',
                        action='store_const', dest='loglevel',
                        const='debug', default='info')
    parser.add_argument('-q', '--quiet', help='set logging to ERROR',
                        action='store_const', dest='loglevel',
                        const='error', default='info')
    parser.add_argument('-w', '--warning', help='set logging to WARNING',
                        action='store_const', dest='loglevel',
                        const='warn', default='info')

    args = parser.parse_args()

    # Create cache directories
    directory_cache = Config.get_default_cache_directory()
    if not os.path.exists(directory_cache):
        print(f'Creating a cache directory at {directory_cache}.')
        os.mkdir(directory_cache)

    # Setup logging.
    #logging.basicConfig(level=args.loglevel,
    #                    format='%(levelname)-8s %(message)s')

    filename_log = os.path.join(directory_cache, 'kaikout_log.csv')
    filename_log_old = os.path.join(directory_cache, 'kaikout_log.csv.old')
    if os.path.exists(filename_log): os.rename(filename_log, filename_log_old)
    loglevel = args.loglevel
    logger = Logger('kaikout')
    if not args.nolog:
        logger.set_filename(filename_log)
        with open(filename_log, 'a') as file:
            file.write('Time,,Level,Module,Function,JID/DB,Message\n')
    logger.set_level(loglevel)

    # Configure settings file
    file_settings = os.path.join(directory_settings, 'accounts.toml')
    if not os.path.exists(file_settings):
        directory_configs = os.path.join(directory, 'configs')
        file_settings_empty = os.path.join(directory_configs, 'accounts.toml')
        shutil.copyfile(file_settings_empty, file_settings)

    data_settings = Toml.open_file(file_settings)

    # Configure account
    data_settings_account = data_settings['xmpp']['client']

    settings_account = {
        'alias': 'Set an Alias',
        'jid': 'Set a Jabber ID',
        'password': 'Input Password'
    }

    for key in settings_account:
        data_settings_account_value = data_settings_account[key]
        if not data_settings_account_value:
            settings_account_message = settings_account[key]
            while not data_settings_account_value:
                if key == 'password':
                    data_settings_account_value = getpass(f'{settings_account_message}: ')
                else:
                    data_settings_account_value = input(f'{settings_account_message}: ')
            data_settings_account[key] = data_settings_account_value

    Toml.save_file(file_settings, data_settings)

    # Try configuration file
    jid = data_settings_account['jid']
    password = data_settings_account['password']
    alias = data_settings_account['alias']
    # TODO
    #hostname = account_xmpp_client['hostname'] if 'hostname' in account_xmpp_client else None
    #port = account_xmpp_client['port'] if 'port' in account_xmpp_client else None
    hostname = port = None
    XmppClient(jid, password, hostname, port, alias)

if __name__ == '__main__':
    main()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import asyncio
import kaikout.config as config
from kaikout.config import Config
from kaikout.log import Logger
from kaikout.database import DatabaseToml
from kaikout.utilities import Documentation, Toml, Url
from kaikout.version import __version__
from kaikout.xmpp.muc import XmppMuc
from kaikout.xmpp.status import XmppStatus
from kaikout.xmpp.utilities import XmppUtilities
from slixmpp.exceptions import IqError, IqTimeout
import sys
import time

try:
    import tomllib
except:
    import tomli as tomllib

logger = Logger(__name__)


    # for task in main_task:
    #     task.cancel()

    # Deprecated in favour of event "presence_available"
    # if not main_task:
    #     await select_file()

class XmppCommands:


    async def clear_filter(self, room, db_file, key):
        function_name = sys._getframe().f_code.co_name
        logger.debug(f'{function_name},{room},Start')
        logger.debug(f'{function_name},{room},{key}')
        value = []
        self.settings[room][key] = value
        DatabaseToml.update_jid_settings(self, room, db_file, key, value)
        message = 'Filter {} has been purged.'.format(key)
        logger.debug(f'{function_name},{room},Finish')
        return message


    async def devoice(self, room, alias, reason):
        function_name = sys._getframe().f_code.co_name
        logger.debug(f'{function_name},{room},Start')
        logger.debug(f'{function_name},{room},{alias}')
        status_message = (
            f'🚫 Committing an action (devoice) against participant {alias}')
        self.action_count += 1
        task_number = self.action_count
        if room not in self.actions: self.actions[room] = {}
        self.actions[room][task_number] = status_message
        XmppStatus.send_status_message(self, room)
        jid_full = XmppMuc.get_full_jid(self, room, alias)
        if jid_full:
            jid_bare = jid_full.split('/')[0]
            await XmppMuc.set_role(self, room, alias, 'visitor', reason)
        await asyncio.sleep(5)
        del self.actions[room][task_number]
        XmppStatus.send_status_message(self, room)
        logger.debug(f'{function_name},{room},Finish')
        return jid_bare


    async def countdown(self, time, room, alias, reason):
        function_name = sys._getframe().f_code.co_name
        logger.debug(f'{function_name},{room},Start')
        logger.debug(f'{function_name},{room},{alias}')
        while time > 1:
            time -= 1
            print([time, room, alias], end='\r')
            await asyncio.sleep(1)
        await XmppCommands.devoice(self, room, alias, reason)
        logger.debug(f'{function_name},{room},Finish')


    def fetch_gemini():
        message = 'Gemini and Gopher are not supported yet.'
        return message


    def get_interval(self, jid_bare):
        result = Config.get_setting_value(
            self.settings, jid_bare, 'interval')
        message = str(result)
        return message


    async def bookmark_add(self, muc_jid):
        if muc_jid not in self.bookmarks: self.bookmarks.append(muc_jid)
        Toml.save_file(self.filename_bookmarks, self.data_bookmarks)
        return f'Groupchat {muc_jid} has been added to bookmarks.'


    async def bookmark_del(self, muc_jid):
        if muc_jid in self.bookmarks: self.bookmarks.remove(muc_jid)
        Toml.save_file(self.filename_bookmarks, self.data_bookmarks)
        return f'Groupchat {muc_jid} has been removed from bookmarks.'


    async def invite_jid_to_muc(self, jid_bare):
        muc_jid = 'slixfeed@chat.woodpeckersnest.space'
        if await XmppUtilities.get_chat_type(self, jid_bare) == 'chat':
            self.plugin['xep_0045'].invite(muc_jid, jid_bare)


    async def kick(self, room, alias, reason):
        function_name = sys._getframe().f_code.co_name
        logger.debug(f'{function_name},{room},Start')
        logger.debug(f'{function_name},{room},{alias}')
        status_message = (
            f'🚫 Committing an action (kick) against participant {alias}')
        self.action_count += 1
        task_number = self.action_count
        if room not in self.actions: self.actions[room] = {}
        self.actions[room][task_number] = status_message
        XmppStatus.send_status_message(self, room)
        jid_full = XmppMuc.get_full_jid(self, room, alias)
        if jid_full:
            jid_bare = jid_full.split('/')[0]
            await XmppMuc.set_affiliation(
                self, room, 'none', jid_bare, reason)
        await asyncio.sleep(5)
        del self.actions[room][task_number]
        XmppStatus.send_status_message(self, room)
        logger.debug(f'{function_name},{room},Finish')
        return jid_bare


    async def muc_join(self, command):
        if command:
            muc_jid = Url.check_xmpp_uri(command)
            if muc_jid:
                # TODO probe JID and confirm it's a group chat
                result = await XmppMuc.join(self, muc_jid)
                if result == 'ban':
                    message = f'{self.alias} is banned from {muc_jid}'
                    if room in self.bookmarks: self.bookmarks.remove(room)
                else:
                    if room not in self.bookmarks: self.bookmarks.append(room)
                    message = f'Joined group chat {muc_jid}'
                Toml.save_file(self.filename_bookmarks, self.data_bookmarks)
            else:
                message = '> {muc_jid}\nGroup chat JID appears to be invalid.'
        else:
            message = '> {}\nGroup chat JID is missing.'
        return message


    async def muc_leave(self, room):
        XmppMuc.leave(self, room)
        if room in self.bookmarks: self.bookmarks.remove(room)
        Toml.save_file(self.filename_bookmarks, self.data_bookmarks)


    async def outcast(self, room, alias, reason):
        function_name = sys._getframe().f_code.co_name
        logger.debug(f'{function_name},{room},Start')
        logger.debug(f'{function_name},{room},{alias}')
        status_message = (
            f'🚫 Committing an action (ban) against participant {alias}')
        self.action_count += 1
        task_number = self.action_count
        if room not in self.actions: self.actions[room] = {}
        self.actions[room][task_number] = status_message
        XmppStatus.send_status_message(self, room)
        jid_full = XmppMuc.get_full_jid(self, room, alias)
        if jid_full:
            jid_bare = jid_full.split('/')[0]
            await XmppMuc.set_affiliation(self, room, 'outcast', jid_bare, reason)
        # else:
        #     # Could "alias" ever be relevant?
        #     # Being a moderator is essential to be able to outcast.
        #     # JIDs are visible to moderators.
        #     await XmppMuc.set_affiliation(self, room, 'outcast', alias, reason)
        await asyncio.sleep(5)
        del self.actions[room][task_number]
        XmppStatus.send_status_message(self, room)
        logger.debug(f'{function_name},{room},Finish')
        return jid_bare


    async def print_bookmarks(self):
        conferences = self.bookmarks
        message = '\nList of groupchats:\n\n```\n'
        for conference in conferences:
            message += f'{conference}\n'
        message += (f'```\nTotal of {len(conferences)} groupchats.\n')
        return message


    def print_help():
        result = Documentation.manual('commands.toml')
        message = '\n'.join(result)
        return message


    def print_help_list():
        command_list = Documentation.manual('commands.toml', section='all')
        message = (
            f'Complete list of commands:\n```\n{command_list}\n```')
        return message


    def print_help_specific(command_root, command_name):
        command_list = Documentation.manual(
            'commands.toml', section=command_root, command=command_name)
        if command_list:
            command_list = ''.join(command_list)
            message = (command_list)
        else:
            message = f'KeyError for {command_root} {command_name}'
        return message


    def print_help_key(command):
        command_list = Documentation.manual('commands.toml', command)
        if command_list:
            command_list = ' '.join(command_list)
            message = (
                f'Available command `{command}` keys:\n```\n{command_list}\n```'
                f'\nUsage: `help {command} <command>`')
        else:
            message = f'KeyError for {command}'
        return message


    def print_info_list():
        config_dir = config.get_default_config_directory()
        with open(config_dir + '/' + 'information.toml', mode="rb") as information:
            result = tomllib.load(information)
        message = '\n'.join(result)
        return message


    def print_info_specific(entry):
        config_dir = config.get_default_config_directory()
        with open(config_dir + '/' + 'information.toml', mode="rb") as information:
            entries = tomllib.load(information)
        if entry in entries:
            # command_list = '\n'.join(command_list)
            message = (entries[entry]['info'])
        else:
            message = f'KeyError for {entry}'
        return message


    def print_options(self, room):
        message = ''
        for key in self.settings[room]:
            val = self.settings[room][key]
            steps = 18 - len(key)
            pulse = ''
            for step in range(steps):
                pulse += ' '
            message += '\n' + key + pulse + ' = ' + str(val)
        return message


    def print_support_jid():
        muc_jid = 'slixfeed@chat.woodpeckersnest.space'
        message = 'Join xmpp:{}?join'.format(muc_jid)
        return message


    def print_unknown():
        message = 'An unknown command.  Type "help" for a list of commands.'
        return message


    def print_version():
        message = __version__
        return message


    def raise_score(self, room, alias, db_file, reason):
        """
        Raise score by one.

        Parameters
        ----------
        room : str
            Jabber ID.
        alias : str
            Alias.
        db_file : str
            Database filename.
        reason : str
            Reason.

        Returns
        -------
        result.

        """
        function_name = sys._getframe().f_code.co_name
        logger.debug(f'{function_name},{room},Start')
        logger.debug(f'{function_name},{room},{alias}')
        status_message = (
            f'✒️ Writing a score against {alias} for {reason}')
        self.action_count += 1
        task_number = self.action_count
        if room not in self.actions: self.actions[room] = {}
        self.actions[room][task_number] = status_message
        XmppStatus.send_status_message(self, room)
        scores = self.settings[room]['scores'] if 'scores' in self.settings[room] else {}
        jid_full = XmppMuc.get_full_jid(self, room, alias)
        if jid_full:
            jid_bare = jid_full.split('/')[0]
            scores[jid_bare] = scores[jid_bare] + 1 if jid_bare in scores else 1
            DatabaseToml.update_jid_settings(
                self, room, db_file, 'scores', scores)
        time.sleep(5)
        del self.actions[room][task_number]
        XmppStatus.send_status_message(self, room)
        result = scores[jid_bare] if jid_full and jid_bare else 0
        logger.debug(f'{function_name},{room},Finish')
        return result


    def update_score_ban(self, room, jid_bare, db_file):
        """
        Update ban score.

        Parameters
        ----------
        room : str
            Jabber ID.
        jid_bare : str
            Jabber ID.
        db_file : str
            Database filename.

        Returns
        -------
        result.

        """
        function_name = sys._getframe().f_code.co_name
        logger.debug(f'{function_name},{room},Start')
        logger.debug(f'{function_name},{room},{jid_bare}')
        scores = self.settings[room]['score_ban'] if 'score_ban' in self.settings[room] else {}
        scores[jid_bare] = scores[jid_bare] + 1 if jid_bare in scores else 1
        DatabaseToml.update_jid_settings(self, room, db_file, 'score_ban', scores)
        logger.debug(f'{function_name},{room},Finish')
        # result = scores[jid_bare]
        # return result


    def update_score_kick(self, room, jid_bare, db_file):
        """
        Update kick score.

        Parameters
        ----------
        room : str
            Jabber ID.
        jid_bare : str
            Jabber ID.
        db_file : str
            Database filename.

        Returns
        -------
        result.

        """
        function_name = sys._getframe().f_code.co_name
        logger.debug(f'{function_name},{room},Start')
        logger.debug(f'{function_name},{room},{jid_bare}')
        scores = self.settings[room]['score_kick'] if 'score_kick' in self.settings[room] else {}
        scores[jid_bare] = scores[jid_bare] + 1 if jid_bare in scores else 1
        DatabaseToml.update_jid_settings(self, room, db_file, 'score_kick', scores)
        logger.debug(f'{function_name},{room},Finish')
        # result = scores[jid_bare]
        # return result


    def update_last_alias(self, room, jid_bare, db_file, alias):
        """
        Update last associated alias.

        Parameters
        ----------
        room : str
            Jabber ID.
        db_file : str
            Database filename.
        jid_bare : str
            Jabber ID.
        alias : 
            Alias.

        Returns
        -------
        None.

        """
        function_name = sys._getframe().f_code.co_name
        logger.debug(f'{function_name},{room},Start')
        logger.debug(f'{function_name},{room},JID: {jid_bare}; Alias: {alias}')
        associated = self.settings[room]['last_alias'] if 'last_alias' in self.settings[room] else {}
        associated[jid_bare] = alias
        DatabaseToml.update_jid_settings(self, room, db_file, 'last_alias', associated)
        logger.debug(f'{function_name},{room},Finish')


    def update_last_activity(self, room, jid_bare, db_file, timestamp):
        """
        Update last message activity.

        Parameters
        ----------
        room : str
            Jabber ID.
        db_file : str
            Database filename.
        jid_bare : str
            Jabber ID.
        timestamp : 
            Time stamp.

        Returns
        -------
        None.

        """
        function_name = sys._getframe().f_code.co_name
        logger.debug(f'{function_name},{room},Start')
        logger.debug(f'{function_name},{room},JID: {jid_bare}; Time: {timestamp}')
        activity = self.settings[room]['last_activity'] if 'last_activity' in self.settings[room] else {}
        activity[jid_bare] = timestamp
        DatabaseToml.update_jid_settings(self, room, db_file, 'last_activity', activity)
        logger.debug(f'{function_name},{room},Finish')


    def remove_last_activity(self, room, jid_bare, db_file):
        """
        Remove last message activity.

        Parameters
        ----------
        room : str
            Jabber ID.
        db_file : str
            Database filename.
        jid_bare : str
            Jabber ID.

        Returns
        -------
        result.

        """
        function_name = sys._getframe().f_code.co_name
        logger.debug(f'{function_name},{room},Start')
        logger.debug(f'{function_name},{room},{jid_bare}')
        activity = self.settings[room]['last_activity'] if 'last_activity' in self.settings[room] else {}
        del activity[jid_bare]
        DatabaseToml.update_jid_settings(self, room, db_file, 'last_activity', activity)
        logger.debug(f'{function_name},{room},Finish')


    def raise_score_inactivity(self, room, alias, db_file):
        """
        Raise score by one.

        Parameters
        ----------
        room : str
            Jabber ID.
        alias : str
            Alias.
        db_file : str
            Database filename.

        Returns
        -------
        result.

        """
        function_name = sys._getframe().f_code.co_name
        logger.debug(f'{function_name},{room},Start')
        logger.debug(f'{function_name},{room},{alias}')
        status_message = (
            f'✒️ Writing a score against {alias} for inactivity')
        self.action_count += 1
        task_number = self.action_count
        if room not in self.actions: self.actions[room] = {}
        self.actions[room][task_number] = status_message
        XmppStatus.send_status_message(self, room)
        scores_inactivity = self.settings[room]['scores_inactivity'] if 'scores_inactivity' in self.settings[room] else {}
        jid_full = XmppMuc.get_full_jid(self, room, alias)
        if jid_full:
            jid_bare = jid_full.split('/')[0]
            scores_inactivity[jid_bare] = scores_inactivity[jid_bare] + 1 if jid_bare in scores_inactivity else 1
            DatabaseToml.update_jid_settings(self, room, db_file, 'scores_inactivity', scores_inactivity)
        time.sleep(5)
        del self.actions[room][task_number]
        XmppStatus.send_status_message(self, room)
        result = scores_inactivity[jid_bare] if jid_full and jid_bare else 0
        logger.debug(f'{function_name},{room},Finish')
        return result


    async def restore_default(self, room, db_file, key=None):
        function_name = sys._getframe().f_code.co_name
        logger.debug(f'{function_name},{room},Start')
        logger.debug(f'{function_name},{room},{key}')
        if key:
            value = self.defaults[key]
            self.settings[room][key] = value
            # data_dir = DatabaseToml.get_default_data_directory()
            # db_file = DatabaseToml.get_data_file(data_dir, jid_bare)
            DatabaseToml.update_jid_settings(self, room, db_file, key, value)
            message = ('Setting {} has been restored to default value.'
                        .format(key))
        else:
            self.settings = self.defaults
            data_dir = DatabaseToml.get_default_data_directory()
            db_file = DatabaseToml.get_data_file(data_dir, room)
            DatabaseToml.create_settings_file(self, db_file)
            message = 'Default settings have been restored.'
        logger.debug(f'{function_name},{room},Finish')
        return message


    def set_filter(self, room, db_file, strings, filter, axis):
        """

        Parameters
        ----------
        db_file : str
            Database filename.
        strings : str
            string (word or phrase).
        filter : str
            'allow' or 'deny'.
        axis : boolean
            True for + (plus) and False for - (minus).

        Returns
        -------
        None.

        """
        function_name = sys._getframe().f_code.co_name
        logger.debug(f'{function_name},{room},Start')
        logger.debug(f'{function_name},{room},{strings}')
        string_list = self.settings[room][filter] if filter in self.settings[room] else []
        new_strings = strings.split(',')
        processed_strings = []
        if axis:
            action = 'added to'
            for string in new_strings:
                if string and string not in string_list:
                    string_trim = string.strip()
                    string_list.append(string_trim)
                    processed_strings.append(string_trim)
        else:
            action = 'removed from'
            for string in new_strings:
                if string and string in string_list:
                    string_trim = string.strip()
                    string_list.remove(string_trim)
                    processed_strings.append(string_trim)
        DatabaseToml.update_jid_settings(self, room, db_file, filter, string_list)
        processed_strings.sort()
        if processed_strings:
            message = 'Strings "{}" have been {} list "{}".'.format(
                ', '.join(processed_strings), action, filter)
        else:
            message = 'Strings "{}" were already {} list "{}".'.format(
                strings, action, filter)
        logger.debug(f'{function_name},{room},Finish')
        return message


    async def set_interval(self, db_file, jid_bare, val):
        try:
            val_new = int(val)
            val_old = Config.get_setting_value(
                self.settings, jid_bare, 'interval')
            await Config.set_setting_value(
                self.settings, jid_bare, db_file, 'interval', val_new)
            message = ('Updates will be sent every {} minutes '
                       '(was: {}).'.format(val_new, val_old))
        except Exception as e:
            logger.error(str(e))
            message = ('No action has been taken.  Enter a numeric value only.')
        return message


    def update_setting_value(self, room, db_file, key, value):
        DatabaseToml.update_jid_settings(self, room, db_file, key, value)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# from slixmpp import JID
from kaikout.database import DatabaseToml
from kaikout.log import Logger
from kaikout.utilities import Documentation
from kaikout.xmpp.commands import XmppCommands
from kaikout.xmpp.message import XmppMessage
from kaikout.xmpp.muc import XmppMuc
from kaikout.xmpp.status import XmppStatus
from kaikout.xmpp.utilities import XmppUtilities
import sys
import time


logger = Logger(__name__)


    # for task in main_task:
    #     task.cancel()

    # Deprecated in favour of event "presence_available"
    # if not main_task:
    #     await select_file()


class XmppChat:


    async def process_message(self, message):
        """
        Process incoming message stanzas. Be aware that this also
        includes MUC messages and error messages. It is usually
        a good practice to check the messages's type before
        processing or sending replies.

        Parameters
        ----------
        message : str
            The received message stanza. See the documentation
            for stanza objects and the Message stanza to see
            how it may be used.
        """

        # jid_bare = message['from'].bare
        # jid_full = str(message['from'])

        # Process commands.
        message_type = message['type']
        message_body = message['body']
        jid = message['from']
        jid_bare = jid.bare

        function_name = sys._getframe().f_code.co_name
        logger.debug(f'{function_name},{jid_bare},Start')

        if (message.get_plugin('muc', check=True) and
            message_type in ('chat', 'groupchat', 'normal')):
            alias = message['muc']['nick']
            room = message['muc']['room']
            self_alias = XmppMuc.get_self_alias(self, jid_bare)
            
            if message_type == 'groupchat':
                if (alias == self.alias or
                    not XmppMuc.is_moderator(self, room, alias) or
                    (not message_body.startswith(self_alias + ' ') and
                     not message_body.startswith(self_alias + ',') and
                     not message_body.startswith(self_alias + '.') and
                     not message_body.startswith(self_alias + ':'))):
                    return
                else:
                    # Adding one to the length because of
                    # assumption that a comma or a dot is added
                    self_alias_length = len(self_alias) + 1
                    command = message_body[self_alias_length:].lstrip()
            elif message_type == 'chat':
                if (alias == self_alias or
                    not XmppMuc.is_moderator(self, jid_bare, alias)):
                    # TODO Count and alert of MUC-PM attempts
                    print(f'Occupant {alias} of group chat {jid_bare} is not approved.')
                    return
                else:
                    command = message_body

        # NOTE Remote session functionality be restored for operators.
        #elif message_type in ('chat', 'normal'):
            """
            command = message_body
            jid_full = jid.full
            room = self.sessions[jid_bare] if jid_bare in self.sessions else message_body
            status_mode = status_text = message_response = None
            # Start a configuration session

            if '@' in room:
                if room in XmppMuc.get_joined_rooms(self):
                    if XmppMuc.is_moderator(self, room, self.alias):
                        is_jid_of_moderators = await XmppUtilities.is_jid_of_moderators(
                            self, room, jid_full)
                        if jid_bare not in self.sessions:
                            if is_jid_of_moderators:
                            # alias = XmppMuc.get_alias(self, room, jid)
                            # if XmppMuc.is_moderator(self, room, alias):
                                room_owners = await XmppMuc.get_affiliation_list(
                                    self, room, 'owner')
                                if not room_owners:
                                    present_participants = XmppMuc.get_roster(
                                        self, room)
                                    present_owners_alias = []
                                    for participant in present_participants:
                                        affiliation = XmppMuc.get_affiliation(
                                            self, room, participant)
                                        if affiliation == 'owner':
                                            present_owners_alias.append(
                                                participant)
                                    present_owners = []
                                    for owner_alias in present_owners_alias:
                                        owner_jid_full = XmppMuc.get_full_jid(
                                            self, room, owner_alias)
                                        owner_jid_bare = owner_jid_full.split('/')[0]
                                        present_owners.append(owner_jid_bare)
                                    if not present_owners:
                                        # present_moderators = set(await XmppMuc.get_role_list(
                                        #     self, room, 'moderator'))
                                        room_admins = await XmppMuc.get_affiliation_list(
                                            self, room, 'admin')
                                # NOTE: Consideration, when there is no access
                                # to the list of owners from groupchat configuration
                                # then send a message to the groupchat - use alias
                                # instead of a Jabber ID.
                                jids_to_notify = room_owners or present_owners or room_admins
                                self.sessions[jid_bare] = room
                                message_response = (
                                    f'A session to configure group chat {room} '
                                    'has been established.')
                                status_mode = 'chat'
                                status_text = f'Session is opened: {room}'
                                for jid in jids_to_notify:
                                    message_notification = (
                                        f'A session for group chat {room} has '
                                        f'been activated by {jid_bare}')
                                    XmppMessage.send(
                                        self, jid, message_notification, 'chat')
                            else:
                                message_response = (
                                    'You do not appear to be a moderator of '
                                    f'group chat {room}')
                                status_mode = 'available'
                                status_text = (
                                    'Type the desired group chat - in which you '
                                    'are a moderator at - to configure')
                                moderators = set(await XmppMuc.get_role_list(
                                    self, room, 'moderator'))
                                message_notification = (
                                    'An unauthorized attempt to establish a '
                                    f'session for group chat {room} has been '
                                    f'made by {jid_bare}')
                                for moderator in moderators:
                                    jid_full = XmppMuc.get_full_jid(
                                        self, room, moderator)
                                    XmppMessage.send(self, jid_full,
                                        message_notification, 'chat')
                        elif is_jid_of_moderators:
                            room = self.sessions[jid_bare]
                        else:
                            del self.sessions[jid_bare]
                            message_response = (
                                'The session has been ended, because you are no '
                                f'longer a moderator at group chat {room}')
                            status_mode = 'away'
                            status_text = f'Session is closed: {room}'
                            moderators = set(await XmppMuc.get_role_list(
                                self, room, 'moderator'))
                            message_notification = (
                                f'The session for group chat {room} with former '
                                f'moderator {jid_bare} has been terminated.\n'
                                f'A termination message was sent to {jid_bare}')
                            for moderator in moderators:
                                jid_full = XmppMuc.get_full_jid(
                                    self, room, moderator)
                                XmppMessage.send(self, jid_full,
                                    message_notification, 'chat')
                    else:
                        message_response = (
                            f'KaikOut must be a moderator at group chat "{room}".')
                else:
                    message_response = (
                        f'KaikOut is not present in group chat {room}'
                        'A session will not begin if KaikOut is not present in '
                        'the desired group chat.\n'
                        f'Invite KaikOut to group chat {room}, and try again')
            else:
                message_response = (
                    f'The text "{room}" does not appear to be a valid '
                    'group chat Jabber ID.')
            if status_mode and status_text:
                XmppStatus.send_status_message(
                    self, jid_full, status_mode, status_text)
            if message_response:
                XmppMessage.send_reply(self, message, message_response)
                return
            """
        else:
            logger.error(
                f'{function_name},{jid_bare},Unknown message type {message_type} {str(message)}')
            return

        db_file = DatabaseToml.instantiate(self, room)

        # # Support private message via groupchat
        # # See https://codeberg.org/poezio/slixmpp/issues/3506
        # if message_type == 'chat' and message.get_plugin('muc', check=True):
        #     # jid_bare = message['from'].bare
        #     if (jid_bare == jid_full[:jid_full.index('/')]):
        #         # TODO Count and alert of MUC-PM attempts
        #         return

        command_time_start = time.time()

        command_lowercase = command.lower()

        # if not self.settings[room]['enabled']:
        #     if not command_lowercase.startswith('start'):
        #         return

        response = None
        if command_lowercase == 'terminate':
            del self.sessions[jid_bare]
            message_response = (
                f'The session for group chat {room} has been terminated.')
            XmppMessage.send_reply(self, message, message_response)
            status_mode = 'away'
            status_text = f'Session is closed: {room}'
            XmppStatus.send_status_message(
                self, jid_full, status_mode, status_text)
        elif command_lowercase.startswith('blacklist +'):
            value = command[11:].strip()
            if value:
                response = XmppCommands.set_filter(
                    self, room, db_file, value, 'jid_blacklist', True)
            else:
                response = 'No action has been taken.  Missing Jabber IDs.'
        elif command_lowercase.startswith('blacklist -'):
            value = command[11:].strip()
            if value:
                response = XmppCommands.set_filter(
                    self, room, db_file, value, 'jid_blacklist', False)
            else:
                response = 'No action has been taken.  Missing Jabber IDs.'
        elif command_lowercase == 'help':
            command_list = XmppCommands.print_help()
            response = ('Available command keys:\n'
                        f'```\n{command_list}\n```\n'
                        'Usage: `help <key>`')
        elif command_lowercase == 'help all':
            command_list = Documentation.manual(
                'commands.toml', section='all')
            response = f'Complete list of commands:\n```\n{command_list}\n```'
        elif command_lowercase.startswith('help'):
            command = command[5:].lower()
            command = command.split(' ')
            if len(command) == 2:
                command_root = command[0]
                command_name = command[1]
                command_list = Documentation.manual(
                    'commands.toml', section=command_root, command=command_name)
                if command_list:
                    command_list = ''.join(command_list)
                    response = (command_list)
                else:
                    response = f'KeyError for {command_root} {command_name}'
            elif len(command) == 1:
                command = command[0]
                command_list = Documentation.manual(
                    'commands.toml', command)
                if command_list:
                    command_list = ' '.join(command_list)
                    response = (f'Available command `{command}` keys:\n'
                                f'```\n{command_list}\n```\n'
                                f'Usage: `help {command} <command>`')
                else:
                    response = f'KeyError for {command}'
            else:
                response = ('Invalid. Enter command key '
                            'or command key & name')
        elif command_lowercase == 'info':
            entries = XmppCommands.print_info_list()
            response = ('Available command options:\n'
                        f'```\n{entries}\n```\n'
                        'Usage: `info <option>`')
        elif command_lowercase.startswith('info'):
            entry = command[5:].lower()
            response = XmppCommands.print_info_specific(entry)
        elif command_lowercase.startswith('action'):
            value = command[6:].strip()
            if value:
                try:
                    value = int(value)
                    XmppCommands.update_setting_value(
                        self, room, db_file, 'action', value)
                    atype = 'ban' if value else 'devoice'
                    response = f'Action has been set to {atype} ({value})'
                except:
                    response = 'Enter a numerical value.'
            else:
                response = str(self.settings[room]['action'])
        elif command_lowercase.startswith('allow +'):
            value = command[7:].strip()
            if value:
                response = XmppCommands.set_filter(
                    self, room, db_file, value, 'allow', True)
            else:
                response = 'No action has been taken.  Missing keywords.'
        elif command_lowercase.startswith('allow -'):
            value = command[7:].strip()
            if value:
                response = XmppCommands.set_filter(
                    self, room, db_file, value, 'allow', False)
            else:
                response = 'No action has been taken.  Missing keywords.'
        elif command_lowercase.startswith('ban'):
            if command[4:]:
                value = command[4:].split()
                alias = value[0]
                reason = ' '.join(value[1:]) if len(value) > 1 else None
                await XmppCommands.outcast(self, room, alias, reason)
        elif command_lowercase.startswith('bookmark +'):
            if XmppUtilities.is_operator(self, jid_bare):
                muc_jid = command[11:]
                response = await XmppCommands.bookmark_add(
                    self, muc_jid)
            else:
                response = 'This action is restricted.  Type: adding bookmarks.'
        elif command_lowercase.startswith('bookmark -'):
            if XmppUtilities.is_operator(self, jid_bare):
                muc_jid = command[11:]
                response = await XmppCommands.bookmark_del(
                    self, muc_jid)
            else:
                response = 'This action is restricted.  Type: removing bookmarks.'
        elif command_lowercase == 'bookmarks':
            if XmppUtilities.is_operator(self, jid_bare):
                response = await XmppCommands.print_bookmarks(self)
            else:
                response = 'This action is restricted.  Type: viewing bookmarks.'
        elif command_lowercase.startswith('clear'):
            key = command[6:]
            response = await XmppCommands.clear_filter(
                self, room, db_file, key)
        elif command_lowercase.startswith('count'):
            value = command[5:].strip()
            if value:
                try:
                    value = int(value)
                    XmppCommands.update_setting_value(
                        self, room, db_file, 'count', value)
                    response = f'Count has been set to {value}'
                except:
                    response = ('No action has been taken.  '
                                'Enter a numerical value.')
            else:
                response = str(self.settings[room]['count'])
        elif command_lowercase.startswith('default'):
            key = command[8:]
            if key:
                response = await XmppCommands.restore_default(
                    self, room, db_file, key)
            else:
                response = 'No action has been taken.  Missing key.'
        elif command_lowercase == 'defaults':
            response = await XmppCommands.restore_default(
                self, room, db_file)
        elif command_lowercase.startswith('deny +'):
            value = command[6:].strip()
            if value:
                response = XmppCommands.set_filter(
                    self, room, db_file, value, 'deny', True)
            else:
                response = 'No action has been taken.  Missing keywords.'
        elif command_lowercase.startswith('deny -'):
            value = command[6:].strip()
            if value:
                response = XmppCommands.set_filter(
                    self, room, db_file, value, 'deny', False)
            else:
                response = 'No action has been taken.  Missing keywords.'
        elif command_lowercase == 'finished off':
            XmppCommands.update_setting_value(
                self, room, db_file, 'finished', 0)
            response = 'Finished indicator has deactivated.'
        elif command_lowercase == 'finished on':
            XmppCommands.update_setting_value(
                self, room, db_file, 'finished', 1)
            response = 'Finished indicator has activated.'
        elif command_lowercase.startswith('frequency messages'):
            value = command[18:].strip()
            if value:
                try:
                    value = int(value)
                    XmppCommands.update_setting_value(
                        self, room, db_file, 'frequency_messages', value)
                    response = ('Minimum allowed frequency for messages '
                                f'has been set to {value}')
                except:
                    response = 'Enter a numerical value.'
            else:
                response = str(self.settings[room]['frequency_messages'])
        elif command_lowercase.startswith('frequency presence'):
            value = command[18:].strip()
            if value:
                try:
                    value = int(value)
                    XmppCommands.update_setting_value(
                        self, room, db_file, 'frequency_presence', value)
                    response = ('Minimum allowed frequency for presence '
                                f'has been set to {value}')
                except:
                    response = 'Enter a numerical value.'
            else:
                response = str(self.settings[room]['frequency_presence'])
        elif command_lowercase == 'exit':
            if message_type == 'groupchat':
                await XmppCommands.muc_leave(self, room)
            else:
                response = 'This command is valid in groupchat only.'
        elif command_lowercase.startswith('ignore +'):
            value = command[8:].strip()
            if value:
                response = XmppCommands.set_filter(
                    self, room, db_file, value, 'rtbl_ignore', True)
            else:
                response = 'No action has been taken.  Missing Jabber IDs.'
        elif command_lowercase.startswith('ignore -'):
            value = command[8:].strip()
            if value:
                response = XmppCommands.set_filter(
                    self, room, db_file, value, 'rtbl_ignore', False)
            else:
                response = 'No action has been taken.  Missing Jabber IDs.'
        elif command_lowercase == 'inactivity off':
            XmppCommands.update_setting_value(
                self, room, db_file, 'check_inactivity', 0)
            response = 'Inactivity check has been deactivated.'
        elif command_lowercase == 'inactivity on':
            XmppCommands.update_setting_value(
                self, room, db_file, 'check_inactivity', 1)
            response = 'Inactivity check has been activated.'
        elif command_lowercase.startswith('inactivity span'):
            value = command[15:].strip()
            if value:
                try:
                    value = int(value)
                    XmppCommands.update_setting_value(
                        self, room, db_file, 'inactivity_span', value)
                    response = ('The maximum allowed time of inactivity '
                                f'has been set to {value} days')
                except:
                    response = 'Enter a numerical value.'
            else:
                response = str(self.settings[room]['inactivity_span'])
        elif command_lowercase.startswith('inactivity warn'):
            value = command[15:].strip()
            if value:
                try:
                    value = int(value)
                    XmppCommands.update_setting_value(
                        self, room, db_file, 'inactivity_warn', value)
                    response = ('The time of inactivity to send a warning '
                                f'upon before action has been set to {value} '
                                'minutes')
                except:
                    response = 'Enter a numerical value.'
            else:
                response = str(self.settings[room]['inactivity_warn'])
        elif command_lowercase.startswith('join'):
            muc_jid = command[5:]
            response = await XmppCommands.muc_join(self, muc_jid)
        elif command_lowercase == 'options':
            response = 'Options:\n```'
            response += XmppCommands.print_options(self, room)
            response += '\n```'
        elif command_lowercase.startswith('kick'):
            if command[5:]:
                value = command[5:].split()
                alias = value[0]
                reason = ' '.join(value[1:]) if len(value) > 1 else None
                await XmppCommands.kick(self, room, alias, reason)
        elif command_lowercase == 'message off':
            XmppCommands.update_setting_value(
                self, room, db_file, 'check_message', 0)
            response = 'Message check has been deactivated.'
        elif command_lowercase == 'message on':
            XmppCommands.update_setting_value(
                self, room, db_file, 'check_message', 1)
            response = 'Message check has been activated.'
        elif command_lowercase.startswith('score messages'):
            value = command[14:].strip()
            if value:
                try:
                    value = int(value)
                    XmppCommands.update_setting_value(
                        self, room, db_file, 'score_messages', value)
                    response = f'Score for messages has been set to {value}'
                except:
                    response = 'Enter a numerical value.'
            else:
                response = str(self.settings[room]['score_messages'])
        elif command_lowercase.startswith('score presence'):
            value = command[14:].strip()
            if value:
                try:
                    value = int(value)
                    XmppCommands.update_setting_value(
                        self, room, db_file, 'score_presence', value)
                    response = f'Score for presence has been set to {value}'
                except:
                    response = 'Enter a numerical value.'
            else:
                response = str(self.settings[room]['score_presence'])
        elif command_lowercase.startswith('scores reset'):
            jid_bare = command[12:].strip()
            if jid_bare:
                if jid_bare in self.settings[room]['scores']:
                    del self.settings[room]['scores'][jid_bare]
                    value = self.settings[room]['scores']
                    XmppCommands.update_setting_value(
                        self, room, db_file, 'scores', value)
                    response = f'Score for {jid_bare} has been reset'
                else:
                    response = f'Jabber ID {jid_bare} is not known.'
            else:
                XmppCommands.update_setting_value(
                    self, room, db_file, 'scores', {})
                response = 'All scores have been reset'
        elif command_lowercase.startswith('scores'):
            jid_bare = command[6:].strip()
            if jid_bare:
                if jid_bare in self.settings[room]['scores']:
                    response = str(self.settings[room]['scores'][jid_bare])
                else:
                    response = f'Jabber ID {jid_bare} is not known.'
            else:
                response = str(self.settings[room]['scores'])
        elif command_lowercase == 'start':
            XmppCommands.update_setting_value(
                self, room, db_file, 'enabled', 1)
            XmppStatus.send_status_message(self, room)
        elif command_lowercase == 'stats':
            response = XmppCommands.print_statistics(db_file)
        elif command_lowercase == 'status off':
            XmppCommands.update_setting_value(
                self, room, db_file, 'check_status', 0)
            response = 'Status message check has been deactivated'
        elif command_lowercase == 'status on':
            XmppCommands.update_setting_value(
                self, room, db_file, 'check_status', 1)
            response = 'Status message check has been activated'
        elif command_lowercase == 'stop':
            XmppCommands.update_setting_value(
                self, room, db_file, 'enabled', 0)
            XmppStatus.send_status_message(self, room)
        elif command_lowercase == 'support':
            response = XmppCommands.print_support_jid()
            await XmppCommands.invite_jid_to_muc(self, room)
        elif command_lowercase.startswith('timer'):
            value = command[5:].strip()
            if value:
                try:
                    value = int(value)
                    XmppCommands.update_setting_value(
                        self, room, db_file, 'timer', value)
                    response = (
                        'Timer value for countdown before committing an action '
                        f'has been set to {value} seconds')
                except:
                    response = 'Enter a numerical value.'
            else:
                response = str(self.settings[room]['timer'])
        elif command_lowercase == 'version':
            response = XmppCommands.print_version()
        elif command_lowercase.startswith('whitelist +'):
            value = command[11:].strip()
            if value:
                response = XmppCommands.set_filter(
                    self, room, db_file, value, 'jid_whitelist', True)
            else:
                response = 'No action has been taken.  Missing Jabber IDs.'
        elif command_lowercase.startswith('whitelist -'):
            value = command[11:].strip()
            if value:
                response = XmppCommands.set_filter(
                    self, room, db_file, value, 'jid_whitelist', False)
            else:
                response = 'No action has been taken.  Missing Jabber IDs.'
        elif command_lowercase.startswith('xmpp:'):
            response = await XmppCommands.muc_join(self, command)
        else:
            response = XmppCommands.print_unknown()

        command_time_finish = time.time()
        command_time_total = command_time_finish - command_time_start
        command_time_total = round(command_time_total, 3)

        if response: XmppMessage.send_reply(self, message, response)

        if room in self.settings and self.settings[room]['finished']:
            response_finished = f'Finished. Total time: {command_time_total}s'
            XmppMessage.send_reply(self, message, response_finished)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

from asyncio import Lock
import csv
from kaikout.log import Logger
from kaikout.utilities import Config, Toml
import os
import sys
import time

# from eliot import start_action, to_file
# # with start_action(action_type="list_feeds()", db=db_file):
# # with start_action(action_type="last_entries()", num=num):
# # with start_action(action_type="get_feeds()"):
# # with start_action(action_type="remove_entry()", source=source):
# # with start_action(action_type="search_entries()", query=query):
# # with start_action(action_type="check_entry()", link=link):

logger = Logger(__name__)


class DatabaseToml:


    def instantiate(self, room):
        """
        Callback function to instantiate action on database.

        Parameters
        ----------
        jid_file : str
            Filename.
        callback : ?
            Function name.
        message : str, optional
            Optional kwarg when a message is a part or
            required argument. The default is None.

        Returns
        -------
        object
            Coroutine object.
        """
        data_dir = DatabaseToml.get_default_data_directory()
        if not os.path.isdir(data_dir):
            os.mkdir(data_dir)
        data_dir_toml = os.path.join(data_dir, 'toml')
        if not os.path.isdir(data_dir_toml):
            os.mkdir(data_dir_toml)
        filename = os.path.join(data_dir_toml, f'{room}.toml')
        if not os.path.exists(filename):
            DatabaseToml.create_settings_file(self, filename)
        DatabaseToml.load_jid_settings(self, room, filename)
        return filename


    def get_default_data_directory():
        if os.environ.get('HOME'):
            data_home = os.path.join(os.environ.get('HOME'), '.local', 'share')
            return os.path.join(data_home, 'kaikout')
        elif sys.platform == 'win32':
            data_home = os.environ.get('APPDATA')
            if data_home is None:
                return 'kaikout_data'
        else:
            return 'kaikout_data'


    def get_data_file(data_dir, room):
        toml_file = os.path.join(data_dir, 'toml', f'{room}.toml')
        return toml_file


    def create_settings_file(self, filename):
        Toml.save_file(filename, self.defaults)


    def load_jid_settings(self, room, filename):
        # data_dir = DatabaseToml.get_default_data_directory()
        # filename = DatabaseToml.get_data_file(data_dir, room)
        self.settings[room] = Toml.open_file(filename)


    def update_jid_settings(self, room, filename, key, value):
        data = Toml.open_file(filename)
        self.settings[room][key] = value
        Toml.save_file(filename, self.settings[room])


class Record:


    def jid_exist(filename, fields):
        """
        Confirm whether Alias and Jabber ID exist.

        Parameters
        ----------
        filename : str
            Filename.
        fields : list
            jid, alias, timestamp.
    
        Returns
        -------
        jid_exist: bool
            False or True.
        """
        jid_exist = False
        data_dir = Config.get_default_data_directory()
        data_dir_logs = os.path.join(data_dir, 'logs')
        csv_file = os.path.join(data_dir_logs, f'{filename}.csv')
        if (not os.path.isdir(data_dir) or
            not os.path.isdir(data_dir_logs) or
            not os.path.exists(csv_file)):
            jid_exist = False
        else:
            with open(csv_file, 'r') as f:
                reader = csv.reader(f)
                for line in reader:
                    if line[0] == fields[0]:
                        jid_exist = True
                        break
        return jid_exist


    def alias_jid_exist(filename, fields):
        """
        Confirm whether Alias and Jabber ID exist.

        Parameters
        ----------
        filename : str
            Filename.
        fields : list
            jid, alias, timestamp.
    
        Returns
        -------
        alias_jid_exist: bool
            False or True.
        """
        alias_jid_exist = False
        data_dir = Config.get_default_data_directory()
        data_dir_logs = os.path.join(data_dir, 'logs')
        csv_file = os.path.join(data_dir_logs, f'{filename}.csv')
        if (not os.path.isdir(data_dir) or
            not os.path.isdir(data_dir_logs) or
            not os.path.exists(csv_file)):
            alias_jid_exist = False
        else:
            with open(csv_file, 'r') as f:
                reader = csv.reader(f)
                for line in reader:
                    if line[0] == fields[0] and line[1] == fields[1]:
                        alias_jid_exist = True
                        break
        return alias_jid_exist


    def csv_jid(filename, fields):
        """
        Log Jabber ID to CSV file.

        Parameters
        ----------
        filename : str
            Filename.
        fields : list
            jid, alias, timestamp.
    
        Returns
        -------
        None.
        """
        data_dir = Config.get_default_data_directory()
        if not os.path.isdir(data_dir): os.mkdir(data_dir)
        data_dir_logs = os.path.join(data_dir, 'logs')
        if not os.path.isdir(data_dir_logs): os.mkdir(data_dir_logs)
        csv_file = os.path.join(data_dir_logs, f'{filename}.csv')
        if not os.path.exists(csv_file):
            columns = ['jid', 'alias', 'timestamp']
            with open(csv_file, 'w') as f:
                writer = csv.writer(f)
                writer.writerow(columns)
        with open(csv_file, 'a') as f:
            writer = csv.writer(f)
            writer.writerow(fields)


    def csv(filename, fields):
        """
        Log message or presence to CSV file.

        Parameters
        ----------
        filename : str
            Filename.
        fields : list
            type, timestamp, alias, body, lang, and identifier.
    
        Returns
        -------
        None.
        """
        data_dir = Config.get_default_data_directory()
        if not os.path.isdir(data_dir): os.mkdir(data_dir)
        data_dir_logs = os.path.join(data_dir, 'logs')
        if not os.path.isdir(data_dir_logs): os.mkdir(data_dir_logs)
        csv_file = os.path.join(data_dir_logs, f'{filename}.csv')
        if not os.path.exists(csv_file):
            columns = ['type', 'timestamp', 'alias', 'body', 'lang', 'identifier']
            with open(csv_file, 'w') as f:
                writer = csv.writer(f)
                writer.writerow(columns)
        with open(csv_file, 'a') as f:
            writer = csv.writer(f)
            writer.writerow(fields)


    def toml(self, room, fields, stanza_type):
        """
        Log message to TOML file.

        Parameters
        ----------
        room : str
            Group chat Jabber ID.
        fields : list
            alias, room, identifier, timestamp.
        stanza_type: str
            message or presence.
    
        Returns
        -------
        None.
        """
        alias, content, identifier, timestamp = fields
        data_dir = DatabaseToml.get_default_data_directory()
        filename = DatabaseToml.get_data_file(data_dir, room)
        # filename = room + '.toml'
        entry = {}
        entry['alias'] = alias
        entry['body'] = content
        entry['id'] = identifier
        entry['timestamp'] = timestamp
        activity_type = 'activity_' + stanza_type
        message_activity_list = self.settings[room][activity_type] if activity_type in self.settings[room] else []
        while len(message_activity_list) > 20: message_activity_list.pop(0)
        message_activity_list.append(entry)
        self.settings[room][activity_type] = message_activity_list # NOTE This directive might not be needed
        Toml.save_file(filename, self.settings[room])





#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""

Functions create_node and create_entry are derived from project atomtopubsub.

"""

import hashlib
from kaikout.log import Logger
from slixmpp.exceptions import IqTimeout, IqError
import sys

logger = Logger(__name__)


class XmppPubsub:


    async def get_pubsub_services(self):
        function_name = sys._getframe().f_code.co_name
        logger.debug(f'{function_name},,Start')
        results = []
        iq = await self['xep_0030'].get_items(jid=self.boundjid.domain)
        items = iq['disco_items']['items']
        for item in items:
            iq = await self['xep_0030'].get_info(jid=item[0])
            identities = iq['disco_info']['identities']
            for identity in identities:
                if identity[0] == 'pubsub' and identity[1] == 'service':
                    result = {}
                    result['jid'] = item[0]
                    if item[1]: result['name'] =  item[1]
                    elif item[2]: result['name'] = item[2]
                    else: result['name'] = item[0]
                    results.extend([result])
        logger.debug(f'{function_name},,Finish')
        return results


    async def get_node_properties(self, jid, node):
        function_name = sys._getframe().f_code.co_name
        logger.debug(f'{function_name},{jid},Start')
        logger.debug(f'{function_name},{jid},{node}')
        config = await self.plugin['xep_0060'].get_node_config(jid, node)
        subscriptions = await self.plugin['xep_0060'].get_node_subscriptions(
            jid, node)
        affiliations = await self.plugin['xep_0060'].get_node_affiliations(
            jid, node)
        properties = {'config': config,
                      'subscriptions': subscriptions,
                      'affiliations': affiliations}
        breakpoint()
        logger.debug(f'{function_name},{jid},Finish')
        return properties


    async def get_node_configuration(self, jid, node_id):
        function_name = sys._getframe().f_code.co_name
        logger.debug(f'{function_name},{jid},Start')
        logger.debug(f'{function_name},{jid},{node_id}')
        node = await self.plugin['xep_0060'].get_node_config(jid, node_id)
        if not node:
            print('NODE CONFIG', node_id, str(node))
            breakpoint()
        logger.debug(f'{function_name},{jid},Finish')
        return node


    async def get_nodes(self, jid):
        function_name = sys._getframe().f_code.co_name
        logger.debug(f'{function_name},{jid},Start')
        nodes = await self.plugin['xep_0060'].get_nodes(jid)
        # 'self' would lead to slixmpp.jid.InvalidJID: idna validation failed:
        logger.debug(f'{function_name},{jid},Finish')
        return nodes


    async def get_item(self, jid, node, item_id):
        function_name = sys._getframe().f_code.co_name
        logger.debug(f'{function_name},{jid},Start')
        logger.debug(f'{function_name},{jid},Node name: {node_id}; Item ID: {item_id}')
        item = await self.plugin['xep_0060'].get_item(jid, node, item_id)
        logger.debug(f'{function_name},{jid},Finish')
        return item


    async def get_items(self, jid, node):
        function_name = sys._getframe().f_code.co_name
        logger.debug(f'{function_name},{jid},Start')
        logger.debug(f'{function_name},{jid},{node}')
        items = await self.plugin['xep_0060'].get_items(jid, node)
        logger.debug(f'{function_name},{jid},Finish')
        return items


    async def subscribe(self, jid, node):
        function_name = sys._getframe().f_code.co_name
        logger.debug(f'{function_name},{jid},Start')
        logger.debug(f'{function_name},{jid},{node}')
        result = await self['xep_0060'].subscribe(jid, node)
        logger.debug(f'{function_name},{jid},Finish')
        return result


    async def get_node_subscriptions(self, jid, node):
        function_name = sys._getframe().f_code.co_name
        logger.debug(f'{function_name},{jid},Start')
        logger.debug(f'{function_name},{jid},{node}')
        try:
            subscriptions = await self['xep_0060'].get_node_subscriptions(
                jid, node)
            return subscriptions
        except IqError as e:
            logger.error(f'{function_name},{jid},Error XmppIQ')
            logger.error(f'{function_name},{jid},"{str(e)}"')
        logger.debug(f'{function_name},{jid},Finish')
